"""Provides a module for generating PyFluent API RST files."""

from pathlib import Path
import shutil


def _get_folder_path(folder_name: str):
    """Get folder path.

    Parameters
    ----------
    folder_name: str
        Name of the folder.

    Returns
    -------
        Path of the folder.
    """
    return (Path(__file__) / ".." / "source" / "api" / folder_name).resolve()


def _get_file_path(folder_name: str, file_name: str):
    """Get file path.

    Parameters
    ----------
    folder_name: str
        Name of the folder.

    file_name: str
        Name of the file.

    Returns
    -------
        Path of the file.
    """
    return (
        Path(__file__) / ".." / "source" / "api" / folder_name / f"{file_name}.rst"
    ).resolve()


hierarchy = {
    "file_reader": ["case_file", "data_file", "lispy"],
    "launcher": [
        "container_launcher",
        "error_handler",
        "fluent_container",
        "launcher_utils",
        "launcher",
        "pim_launcher",
        "process_launch_string",
        "pyfluent_enums",
        "slurm_launcher",
        "standalone_launcher",
        "watchdog",
    ],
    "meshing": ["meshing_workflow", "datamodel/datamodel_contents", "tui/tui_contents"],
    "post_objects": [
        "check_in_notebook",
        "post_helper",
        "post_objects_container",
        "singleton_meta",
        "timing_decorator",
    ],
    "scheduler": ["load_machines", "machine_list"],
    "services": [
        "api_upgrade",
        "batch_ops",
        "datamodel_se",
        "datamodel_tui",
        "events",
        "field_data",
        "health_check",
        "interceptors",
        "monitor",
        "reduction",
        "scheme_eval",
        "settings",
        "solution_variables",
        "streaming",
        "transcript",
    ],
    "solver": [
        "error_message",
        "flobject",
        "datamodel/datamodel_contents",
        "settings_root",
        "tui/tui_contents",
    ],
    "streaming_services": [
        "datamodel_event_streaming",
        "datamodel_streaming",
        "events_streaming",
        "field_data_streaming",
        "monitor_streaming",
        "streaming",
        "transcript_streaming",
    ],
    "utils": [
        "data_transfer",
        "deprecate",
        "dictionary_operations",
        "dump_session_data",
        "event_loop",
        "execution",
        "file_transfer_service",
        "fix_doc",
        "fldoc",
        "fluent_version",
        "networking",
        "setup_for_fluent",
    ],
    "other": [
        "exceptions",
        "file_session",
        "fluent_connection",
        "journaling",
        "logging",
        "parametric",
        "rpvars",
        "search",
        "session_base_meshing",
        "session_meshing",
        "session_pure_meshing",
        "session_solver_icing",
        "session_solver_lite",
        "session_solver",
        "session",
        "system_coupling",
        "warnings",
        "workflow",
    ],
}


def _write_common_rst_members(rst_file):
    rst_file.write("    :members:\n")
    rst_file.write("    :show-inheritance:\n")
    rst_file.write("    :undoc-members:\n")
    rst_file.write("    :exclude-members: __weakref__, __dict__\n")
    rst_file.write("    :special-members: __init__\n")
    rst_file.write("    :autosummary:\n")


def _generate_api_source_rst_files(folder: str, files: list):
    for file in files:
        if file.endswith("_contents"):
            pass
        else:
            if folder:
                rst_file = _get_file_path(folder, file)
            else:
                rst_file = _get_file_path("", file)
            with open(rst_file, "w", encoding="utf8") as rst:
                rst.write(f".. _ref_{file}:\n\n")
                if folder:
                    if "root" in file:
                        rst.write("solver.settings\n")
                        rst.write(f'{"="*(len("solver.settings"))}\n\n')
                        rst.write(
                            "The :ref:`ref_root` is the top-level solver settings object. It contains all\n"
                        )
                        rst.write(
                            "other settings objects in a hierarchical structure.\n"
                        )
                    else:
                        rst.write(f"{file}\n")
                        rst.write(f'{"="*(len(f"{file}"))}\n\n')
                        rst.write(
                            f".. automodule:: ansys.fluent.core.{folder}.{file}\n"
                        )
                else:
                    rst.write(f"{file}\n")
                    rst.write(f'{"="*(len(f"{file}"))}\n\n')
                    rst.write(f".. automodule:: ansys.fluent.core.{file}\n")
                if "root" not in file:
                    _write_common_rst_members(rst_file=rst)


def _generate_api_index_rst_files():
    for folder, files in hierarchy.items():
        if Path(_get_folder_path(folder)).is_dir():
            shutil.rmtree(_get_folder_path(folder))
        if folder == "other":
            _generate_api_source_rst_files(None, files)
        else:
            Path(_get_folder_path(folder)).mkdir(parents=True, exist_ok=True)
            folder_index = _get_file_path(folder, f"{folder}_contents")
            with open(folder_index, "w", encoding="utf8") as index:
                index.write(f".. _ref_{folder}:\n\n")
                index.write(f"{folder}\n")
                index.write(f'{"="*(len(f"{folder}"))}\n\n')
                index.write(f".. automodule:: ansys.fluent.core.{folder}\n")
                _write_common_rst_members(rst_file=index)
                index.write(".. toctree::\n")
                index.write("    :maxdepth: 2\n")
                index.write("    :hidden:\n\n")
                for file in files:
                    index.write(f"    {file}\n")
                index.write("\n")
            _generate_api_source_rst_files(folder, files)


if __name__ == "__main__":
    _generate_api_index_rst_files()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_lunar_lander_thermal:

Thermal Model of a Lunar Lander Using the Monte Carlo Radiation Model
---------------------------------------------------------------------
This example demonstrates creating and solving a thermal model of a lander on
the lunar surface using Fluent's Monte Carlo radiation solver.

PyFluent uses the following loop at each timestep to retrofit the required
functionality to Fluent:

1. Calculate the current sun vector (direction of Sun relative to observer).
2. Update the radiation direction in Fluent.
3. Change the radiator emissivity depending on whether the louvers are open.
4. Run the solution for 1 timestep.

**Workflow tasks**

The Thermal Model of a Lunar Lander Using the Monte Carlo Radiation Model
example guides you through these tasks:

* Setting up a Monte Carlo radiation model.
* Creation of materials with thermal and radiation properties.
* Setting boundary conditions for heat transfer and radiation calculations.
* Setting up shell conduction boundary conditions.
* Calculating a solution using the pressure-based solver.
* Dynamically updating the Sun direction and lander state at each step.

**Problem description**

The lander is modelled as a hollow 1 m × 1 m × 1 m cube with aluminum walls 3
mm thick, covered in highly reflective multilayer insulation (MLI). To allow
for comparison to empirical data, the landing site is selected to be the same
as that of the Apollo 17 mission, which took measurements of the regolith
temperatures in the Taurus-Littrow valley located at 20.1908° N, 30.7717° E.

The lander has a radiator on the top surface of size 0.7 m × 0.7 m with louvers,
which open above 273 K to reject heat and close below 273 K to retain heat.
The PyFluent API is used to automatically update the radiator state and the
direction of the Sun at each timestep.

The case setup is taken from reference [1_], originally created in ANSYS
Thermal Desktop. It uses a thermal model of the lunar regolith developed in
reference [2_]. Validation data for the regolith temperatures are taken from
measurements conducted by the Apollo 17 mission to the Moon [3_].
"""

# sphinx_gallery_thumbnail_path = '_static/lunar_lander_thermal_thumbnail.png'

###############################################################################
# .. image:: ../../_static/lunar_lander_thermal_setup.png
#   :align: center

###############################################################################
# Preliminary Setup
# -----------------
# First, we will define functions to compute the direction of the Sun relative
# to the lander.
#
# Perform required imports
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Perform required imports, including downloading the required geometry files.
# The mesh has been pre-made for this problem.

# flake8: noqa: E402

from itertools import chain

import numpy as np

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples

lander_spaceclaim_file, lander_mesh_file, apollo17_temp_data = [
    examples.download_file(
        f,
        "pyfluent/lunar_lander_thermal",
    )
    for f in [
        "lander_geom.scdoc",
        "lander_mesh.msh.h5",
        "apollo17_temp_data.csv",
    ]
]

###############################################################################
# Define variables
# ~~~~~~~~~~~~~~~~
# Define the key variables:
#
# * The obliquity (axial tilt) of the Moon relative to the plane of the Earth's
#   orbit around the Sun
# * The setpoint temperature above which the radiator louvers open
# * The coordinates of the lander
# * The timestep size
# * The number of timesteps
#
# We will use a timestep size of 24 hours and run the simulation for 60 Earth
# days, or approximately 2 lunar days.

# Lunar axial obliquity relative to ecliptic
moon_obliquity = np.deg2rad(1.54)

# Louver setpoint temperature [K]
louver_setpoint_temp = 273

# Lander coordinates (Apollo 17 landing site)
land_lat = np.deg2rad(20.1908)
land_lon = np.deg2rad(30.7717)

# Timestep size of 1 Earth day
step_size = 86400

# Run simulation for 60 Earth days
n_steps = 60

###############################################################################
# Define sun vector function
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Define the function to calculate the Sun vector at the landing site. This is
# a vector that points from the observer's location toward the Sun. It takes
# Earth's ecliptic longitude at the beginning of the simulation, the subsolar
# longitude at the beginning of the simulation, the observer's coordinates, and
# the time elapsed in seconds since the beginning of the simulation as inputs.
# It returns the Sun's altitude and azimuth as outputs.
#
# For more information, see the following links:
#
# * `Ecliptic longitude <https://w.wiki/7zFR>`_
# * `Subsolar longitude <https://w.wiki/7zFS>`_
# * `Altitude and azimuth <https://w.wiki/7zFU>`_


def calc_sun_vecs_for_moon(
    earth_ecliptic_lon_start: float,
    subsol_lon_start: float,
    obsv_lat: float,
    obsv_lon: float,
    t: float,
) -> tuple[float, float]:
    """Calculate sun vectors."""
    # Earth ecliptic longitude
    earth_ecliptic_lon = earth_ecliptic_lon_start + 2 * np.pi / (365 * 86400) * t

    # Subsolar point
    subsol_lat = np.arcsin(np.sin(-moon_obliquity) * np.sin(earth_ecliptic_lon))
    subsol_lon = subsol_lon_start + t / (29.5 * 86400) * 2 * np.pi

    # Solar altitude and azimuth
    sun_obsv_phaseang = np.arccos(
        np.sin(subsol_lat) * np.sin(obsv_lat)
        + np.cos(subsol_lat) * np.cos(obsv_lat) * np.cos(obsv_lon - subsol_lon)
    )

    sun_alt = np.pi / 2 - sun_obsv_phaseang
    sun_azm = np.arctan2(
        np.cos(obsv_lat) * np.sin(subsol_lat)
        - np.sin(obsv_lat) * np.cos(subsol_lat) * np.cos(subsol_lon - obsv_lon),
        np.cos(subsol_lat) * np.sin(subsol_lon - obsv_lon),
    )

    return sun_alt, sun_azm


###############################################################################
# Define beam direction function
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Define the function that converts the Sun's altitude and azimuth, as
# calculated by ``calc_sun_vecs_for_moon``, into a beam direction in Cartesian
# coordinates that can be used in Fluent. The coordinate system used is:
#
# * X: North
# * Y: Zenith
# * Z: East


def sun_vec_to_beam_dir(
    sun_alt: float,
    sun_azm: float,
) -> tuple[float, float, float]:
    """Calculate beam direction."""
    # Coordinate system:
    #   X: North
    #   Y: Zenith
    #   Z: East
    x = np.cos(sun_alt) * np.cos(sun_azm)
    y = np.sin(sun_alt)
    z = np.cos(sun_alt) * np.sin(sun_azm)

    # Since the sun vector points toward the sun, we need to take the
    # negative to get the beam direction
    return -x, -y, -z


###############################################################################
# Define mean surface temperature function
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# The radiator will be opened or closed based on its mean temperature. We will
# access the solver session object to extract nodal temperatures. It takes a
# list of surface names as input, finds their surface IDs, obtains the scalar
# field data from the solver, then returns the average temperature.


def get_surf_mean_temp(
    surf_names: list[str],
    solver: pyfluent.session_solver.Solver,
) -> float:
    """Calculate mean surface temperature."""
    # Get surface IDs
    surfs = solver.field_info.get_surfaces_info()
    surf_ids_ = [surfs[surf_name]["surface_id"] for surf_name in surf_names]

    # Flatten surf_ids nested list
    surf_ids = list(chain(*surf_ids_))

    # Get temperature data
    temp_data = solver.field_data.get_scalar_field_data(
        "temperature",
        surfaces=surf_ids,
    )

    # Calculate mean temperature across surfaces
    temps = np.array([])
    for x in temp_data.values():
        temps = np.concatenate(
            (
                temps,
                np.array([y.scalar_data for y in x.data]),
            ),
        )

    return np.mean(temps)


###############################################################################
# Start Fluent
# ------------
# We are now ready to launch Fluent and load the mesh.
#
# Launch Fluent and print Fluent version
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

solver = pyfluent.launch_fluent(
    precision="double",
    processor_count=12,
    mode="solver",
    cwd=pyfluent.EXAMPLES_PATH,
    product_version="25.1.0",
)
print(solver.get_fluent_version())

###############################################################################
# Load the mesh
# ~~~~~~~~~~~~~

solver.file.read_mesh(file_name=lander_mesh_file)

###############################################################################
# Case Setup
# ----------
# We are now ready to begin setting up the case.
#
# Transient settings
# ~~~~~~~~~~~~~~~~~~
# Set the solution type to transient and configure. We will use 2nd-order time-
# stepping, a fixed timestep size, and a limit of 20 iterations per timestep.
# Since we are only running 1 timestep at a time, the time step count is set to
# 1.

# Set solution to transient
solver.setup.general.solver.time = "unsteady-2nd-order"

# Set transient settings
trans_controls = solver.solution.run_calculation.transient_controls
trans_controls.type = "Fixed"
trans_controls.max_iter_per_time_step = 20
trans_controls.time_step_count = 1
trans_controls.time_step_size = step_size

###############################################################################
# Enable models
# ~~~~~~~~~~~~~
# Enable the energy model. Since fluid flow is not simulated, we will set the
# viscosity model to laminar.

models = solver.setup.models
models.energy.enabled = True
models.viscous.model = "laminar"

###############################################################################
# Set up radiation model
# ~~~~~~~~~~~~~~~~~~~~~~
# Enable the Monte Carlo radiation model with two radiation bands: one for
# solar radiation and one for thermal infrared radiation. Ensure that bands are
# created in order of increasing wavelength.
#
# The number of histories is set to 10 million to reduce computation time, but
# more may be required for accurate results.
#
# The limits of each band are based on Fluent manual recommendations and on
# space industry best practices [4_].


# Set up radiation model
radiation = models.radiation
radiation.model = "monte-carlo"
radiation.monte_carlo.number_of_histories = 1e7

# Define range of solar wavelengths
radiation.multiband["solar"] = {
    "start": 0,
    "end": 2.8,
}
# Define range of thermal IR wavelengths
radiation.multiband["thermal-ir"] = {
    "start": 2.8,
    "end": 100,
}

# Solve radiation once per timestep
radiation_freq = radiation.solve_frequency
radiation_freq.method = "time-step"
radiation_freq.time_step_interval = 1

###############################################################################
# Define materials
# ~~~~~~~~~~~~~~~~
# Create materials to represent the vacuum of space (to prevent thermal
# conduction through the cell zone representing vacuum) and the lunar regolith
# (soil). The thermal conductivity of the regolith and the surface 'fluff' is
# strongly temperature-dependent and so must be modelled using an expression.

# --- Properties of vacuum ---
# Thermal conductivity: 0

vacuum = solver.setup.materials.solid.create("vacuum")
vacuum.chemical_formula = ""
vacuum.thermal_conductivity.value = 0
vacuum.absorption_coefficient.value = 0
vacuum.refractive_index.value = 1

# --- Properties of fluff (see ref. [2]) ---
# Density: 1000 [kg m^-3]
# Specific heat capacity: 1050 [J kg^-1 K^-1]
# Thermal conductivity: 9.22e-4*(1 + 1.48*(temperature/350 K)^3) [W m^-1 K^-1]

fluff = solver.setup.materials.solid.create("fluff")
fluff.chemical_formula = ""
fluff.density.value = 1000
fluff.specific_heat.value = 1050
fluff.thermal_conductivity.option = "expression"
fluff.thermal_conductivity.expression = (
    "9.22e-4[W m^-1 K^-1]*(1 + 1.48*(StaticTemperature/350[K])^3)"
)

# --- Properties of regolith (see ref. [2]) ---
# Density: 2000 [kg m^-3]
# Specific heat capacity: 1050 [J kg^-1 K^-1]
# Thermal conductivity: 9.30e-4*(1 + 0.73*(temperature/350 K)^3) [W m^-1 K^-1]

regolith = solver.setup.materials.solid.create("regolith")
regolith.chemical_formula = ""
regolith.density.value = 2000
regolith.specific_heat.value = 1050
regolith.thermal_conductivity.option = "expression"
regolith.thermal_conductivity.expression = (
    "9.30e-4[W m^-1 K^-1]*(1 + 0.73*(StaticTemperature/350[K])^3)"
)

###############################################################################
# Cell zone conditions
# ~~~~~~~~~~~~~~~~~~~~
# Since the Monte Carlo radiation model only supports radiative transfer
# through a cell zone, a cell zone is used to model the vacuum around the
# lander. This cell zone must be set to be a solid so that the fluid equations
# are not solved there, then it must be assigned to the vacuum material.

cellzones = solver.setup.cell_zone_conditions
cellzones.set_zone_type(
    zone_list=["geom-2_domain"],
    new_type="solid",
)
cellzones.solid["geom-2_domain"].material = "vacuum"

###############################################################################
# Regolith boundary condition
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# We will implement the regolith model in reference [2_] by modelling
# it as a wall exhibiting multilayer shell conduction. A basal heat flux is
# used to represent the geothermal heat from the Moon's interior that heats
# the regolith from the bottom.

# --- Regolith BC ---
# Thickness of layers: 0.02, 0.04, 0.08, 0.16, 0.32 [m]
# Heating at base: 0.031 [W m^-2]
# Surface absorptivity: 0.87
# Surface emissivity: 0.97

regolith_bc = solver.setup.boundary_conditions.wall["regolith"]

regolith_bc.thermal.q.value = 0.031
regolith_bc.thermal.planar_conduction = True
regolith_bc.thermal.shell_conduction = [
    {
        "thickness": 0.02,
        "material": "fluff",
    },
    {
        "thickness": 0.04,
        "material": "regolith",
    },
    {
        "thickness": 0.08,
        "material": "regolith",
    },
    {
        "thickness": 0.16,
        "material": "regolith",
    },
    {
        "thickness": 0.32,
        "material": "regolith",
    },
]
regolith_bc.radiation.band_in_emiss = {
    "solar": 0.87,
    "thermal-ir": 0.97,
}

###############################################################################
# Space boundary condition
# ~~~~~~~~~~~~~~~~~~~~~~~~
# The space boundary condition represents deep space and also acts as the
# source of the Sun's illumination in the simulation.

# --- Set up space boundary condition ---
# Temperature: 3 [K]
# Emissivity: 1
# Absorptivity: 1
# Solar flux: 1414 [W m^-2]

space_bc = solver.setup.boundary_conditions.wall["space"]

space_bc.thermal.thermal_bc = "Temperature"
space_bc.thermal.t.value = 3
space_bc.thermal.material = "vacuum"
space_bc.radiation.mc_bsource_p = True
space_bc.radiation.band_q_irrad["solar"].value = 1414
space_bc.radiation.band_diffuse_frac = {
    "solar": 0,
    "thermal-ir": 0,
}
space_bc.radiation.band_in_emiss = {
    "solar": 1,
    "thermal-ir": 1,
}

###############################################################################
# Spacecraft walls boundary condition
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# The spacecraft is covered in reflective MLI (multilayer insulation) and heat
# transfer through the aluminum shell can be simulated using shell conduction.

# --- Set up spacecraft shell boundary condition ---
# Thickness: 0.03 [m]
# Material: aluminum
# Absorptivity: 0.05
# Emissivity: 0.05

sc_mli_bc = solver.setup.boundary_conditions.wall["sc-mli"]

sc_mli_bc.thermal.planar_conduction = True
sc_mli_bc.thermal.shell_conduction = [
    {
        "thickness": 0.03,
        "material": "aluminum",
    },
]
sc_mli_bc.radiation.band_in_emiss = {
    "solar": 0.05,
    "thermal-ir": 0.05,
}

###############################################################################
# Spacecraft radiator boundary condition
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# The spacecraft's radiator is modelled in a similar way to the spacecraft's
# walls, but the emissivity is left unset as it will be changed dynamically by
# our PyFluent script depending on its temperature during the simulation.

# --- Set up spacecraft radiator boundary condition ---
# Thickness: 0.03 [m]
# Material: aluminum
# Absorptivity: 0.17
# Emissivity: 0.09 below 273 K, 0.70 otherwise

sc_rad_bc = solver.setup.boundary_conditions.wall["sc-radiator"]

sc_rad_bc.thermal.planar_conduction = True
sc_rad_bc.thermal.shell_conduction = [
    {
        "thickness": 0.03,
        "material": "aluminum",
    },
]
sc_rad_bc.radiation.band_in_emiss = {
    "solar": 0.17,
}

###############################################################################
# Initialize simulation
# ~~~~~~~~~~~~~~~~~~~~~
# Initialize the simulation. Doing so now will create conductive zones between
# shell layers representing the regolith, allowing us to make report
# definitions on them in the next step. The entire domain will be initialized
# to a temperature of 230 K, or -43 °C.

sim_init = solver.solution.initialization
sim_init.defaults["temperature"] = 230
sim_init.initialize()

###############################################################################
# Spacecraft temperature reports
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Create reports for the spacecraft's minimum, mean, and maximum temperatures.

surf_report_defs = solver.solution.report_definitions.surface

sc_surfs = ["sc-radiator", "sc-mli"]

surf_report_defs["sc-min-temp"] = {
    "surface_names": sc_surfs,
    "report_type": "surface-facetmin",
    "field": "temperature",
}
surf_report_defs["sc-avg-temp"] = {
    "surface_names": sc_surfs,
    "report_type": "surface-facetavg",
    "field": "temperature",
}
surf_report_defs["sc-max-temp"] = {
    "surface_names": sc_surfs,
    "report_type": "surface-facetmax",
    "field": "temperature",
}

###############################################################################
# Regolith temperature reports
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Create reports for the mean temperatures of each regolith layer. We will
# store the report names for use later.

surf_report_defs = solver.solution.report_definitions.surface

# Loop over all regolith reports to set common properties
regolith_report_names = []
for i in range(1, 5 + 1):
    report_name = f"regolith-layer-{i}-temp"
    surf_report_defs[report_name] = {
        "report_type": "surface-facetavg",
        "field": "temperature",
    }
    regolith_report_names.extend([report_name])

surf_report_defs["regolith-layer-1-temp"].surface_names = ["regolith"]
surf_report_defs["regolith-layer-2-temp"].surface_names = ["regolith-1:2"]
surf_report_defs["regolith-layer-3-temp"].surface_names = ["regolith-2:3"]
surf_report_defs["regolith-layer-4-temp"].surface_names = ["regolith-3:4"]
surf_report_defs["regolith-layer-5-temp"].surface_names = ["regolith-4:5"]

###############################################################################
# Temperature report files
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Create temperature report files for post-processing.

surf_report_files = solver.solution.monitor.report_files

# Spacecraft temperatures
surf_report_files["sc-temps-rfile"] = {
    "report_defs": ["flow-time", "sc-min-temp", "sc-avg-temp", "sc-max-temp"],
}

# Regolith temperatures
surf_report_files["regolith-temps-rfile"] = {
    "report_defs": [*regolith_report_names, "flow-time"],
}

###############################################################################
# Autosave
# ~~~~~~~~
# Set the case to save only the data file at each timestep for post-processing.

autosave = solver.file.auto_save

autosave.case_frequency = "if-mesh-is-modified"
autosave.data_frequency = 1
autosave.save_data_file_every.frequency_type = "time-step"
autosave.append_file_name_with.file_suffix_type = "time-step"

###############################################################################
# Convergence criteria
# ~~~~~~~~~~~~~~~~~~~~
# Turn off the convergence criteria pertaining to fluid flow as there is no
# fluid flow in this simulation. Keep only the energy convergence criterion.

residuals = solver.solution.monitor.residual.equations

for criterion in ["continuity", "x-velocity", "y-velocity", "z-velocity"]:
    residuals[criterion].check_convergence = False
    residuals[criterion].monitor = False

###############################################################################
# Write case file
# ~~~~~~~~~~~~~~~
# Write the case file. Enable overwrite.

solver.file.batch_options.confirm_overwrite = True
solver.file.write(
    file_name="lunar_lander_thermal.cas.h5",
    file_type="case",
)

###############################################################################
# Solve
# -----
#
# Solve case
# ~~~~~~~~~~
# Run the case, using a loop to update the radiator state and Sun vector at
# each timestep. We will assume that the Earth's ecliptic longitude and the
# subsolar longitude are both zero at the start of the simulation for
# simplicity.

for i in range(n_steps):
    # Get current simulation time
    t = solver.rp_vars("flow-time")

    # Calculate sun vector
    sun_alt, sun_azm = calc_sun_vecs_for_moon(
        earth_ecliptic_lon_start=0,
        subsol_lon_start=0,
        obsv_lat=land_lat,
        obsv_lon=land_lon,
        t=t,
    )
    beam_x, beam_y, beam_z = sun_vec_to_beam_dir(
        sun_alt=sun_alt,
        sun_azm=sun_azm,
    )

    # Set beam direction
    solver.setup.boundary_conditions.wall["space"].radiation.reference_direction = [
        beam_x,
        beam_y,
        beam_z,
    ]

    # Calculate radiator mean temperature
    rad_mean_temp = get_surf_mean_temp(
        ["sc-radiator"],
        solver,
    )

    # Simulate closing louvers below 273 K by changing emissivity
    rad_emiss = solver.setup.boundary_conditions.wall[
        "sc-radiator"
    ].radiation.band_in_emiss["thermal-ir"]
    if rad_mean_temp < 273:
        rad_emiss.value = 0.09
    else:
        rad_emiss.value = 0.70

    # Run simulation for 1 timestep
    solver.solution.run_calculation.calculate()

###############################################################################
# Close Fluent
# ~~~~~~~~~~~~
# Shut down the solver.

solver.exit()

###############################################################################
# Post-process
# ------------
# Post-process the data.
#
# Import packages
# ~~~~~~~~~~~~~~~
# Import the packages required for post-processing.

from pathlib import Path
import re

from matplotlib import pyplot as plt
import pandas as pd

###############################################################################
# Clean column names function
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Pandas will not correctly format the column names when reading in the Fluent
# output files. We will clean the names by removing all parentheses and double
# quotation marks.


def clean_col_names(df):
    """Clean column names."""
    df.columns = [re.sub(r'["()]', "", col) for col in df.columns]


###############################################################################
# Read in simulation data
# ~~~~~~~~~~~~~~~~~~~~~~~
# Read the Fluent output files for the regolith and spacecraft temperature data
# into Pandas dataframes. The separation delimiter between columns is defined
# as one or more spaces that is not immediately in front of or behind an
# alphabetical character, implemented as negative lookarounds in a regular
# expression.

root = Path(pyfluent.EXAMPLES_PATH)
sep = r"(?<![a-zA-Z])\s+(?![a-zA-Z])"

# Read in regolith data
regolith_df = pd.read_csv(
    root / "regolith-temps-rfile.out",
    sep=sep,
    header=2,
    dtype=np.float64,
    engine="python",
)
clean_col_names(regolith_df)

# Read in spacecraft data
sc_df = pd.read_csv(
    root / "sc-temps-rfile.out",
    sep=sep,
    header=2,
    dtype=np.float64,
    engine="python",
)
clean_col_names(sc_df)

###############################################################################
# Read in experimental data
# ~~~~~~~~~~~~~~~~~~~~~~~~~
# Read in the experimental data from Apollo 17 [3_] to compare against the
# simulation. We will offset the data to start at time step 25 of the
# simulation, to ensure that the data are compared at the same sun angles.

apollo17_df = pd.read_csv(apollo17_temp_data)

apollo17_offset = regolith_df["flow-time"].iloc[25]
apollo17_df["Offset time since sunrise"] = (
    apollo17_df["Time since sunrise"] + apollo17_offset
)

###############################################################################
# Set data types
# ~~~~~~~~~~~~~~
# Set the data type of the time step column to integer.

regolith_df = regolith_df.astype(
    {
        "Time Step": np.int64,
    },
)
sc_df = sc_df.astype(
    {
        "Time Step": np.int64,
    },
)

###############################################################################
# Plot regolith temperatures
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Plot the mean temperatures of each layer of regolith, as well as a comparison
# with the Apollo 17 data. The simulation agrees well with experiment.

fig1, ax1 = plt.subplots()

regolith_df.plot(
    x="flow-time",
    y=[f"regolith-layer-{i+1}-temp" for i in range(5)],
    title="Regolith temperatures",
    ax=ax1,
)
apollo17_df.plot(
    x="Offset time since sunrise",
    y="Temperature",
    style="x",
    xlabel="Time [s]",
    ylabel="Temperature [K]",
    ax=ax1,
)

ax1.legend(["Layer 1", "2", "3", "4", "5", "Apollo 17"])

###############################################################################
# Plot spacecraft temperatures
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Plot the minimum, mean, and maximum temperatures of the spacecraft.

fig2, ax2 = plt.subplots()

sc_df.plot(
    x="flow-time",
    y=["sc-min-temp", "sc-avg-temp", "sc-max-temp"],
    title="Spacecraft Temperatures",
    xlabel="Time [s]",
    ylabel="Temperature [K]",
    ax=ax2,
)
ax2.axhline(273, color="k", linestyle=":")

ax2.legend(["Minimum", "Mean", "Maximum", "Setpoint"])

###############################################################################
# Show plots
# ~~~~~~~~~~
# Display the plots.

plt.show()

###############################################################################
# The expected plot for the mean temperature of the regolith layers (Figure 1)
# is as follows. This represents the regolith temperatures that would be
# observed without a spacecraft in the environment.
#
# .. image:: ../../_static/lunar_lander_regolith_temps.png
#   :align: center
#
# The expected plot for the spacecraft's minimum, mean, and maximum
# temperatures (Figure 2) is as follows.
#
# .. image:: ../../_static/lunar_lander_sc_temps.png
#   :align: center


###############################################################################
# References
# ------------
#
# .. _1:
#
# [1] T.-Y. Park, J.-J. Lee, J.-H. Kim, and H.-U. Oh, “Preliminary Thermal
# Design and Analysis of Lunar Lander for Night Survival,” *International
# Journal of Aerospace Engineering*, vol. 2018, p. e4236396, Oct. 2018, doi:
# `doi.org/10.1155/2018/4236396 <https://doi.org/10.1155/2018/4236396>`_.
#
# .. _2:
#
# [2] R. J. Christie, D. W. Plachta, and M. M. Yasan, “Transient Thermal Model
# and Analysis of the Lunar Surface and Regolith for Cryogenic Fluid Storage,”
# NASA Glenn Research Center, Cleveland, Ohio, NASA Technical Report
# TM-2008-215300, Aug. 2008. [Online]. Available:
# https://ntrs.nasa.gov/citations/20080039640
#
# .. _3:
#
# [3] M. G. Langseth, S. J. Keihm, and J. L. Chute, “Heat Flow Experiment,” in
# *Apollo 17: Preliminary Science Report*, vol. SP-330, Washington, D.C.: NASA
# Lyndon B. Johnson Space Center, 1973. [Online]. Available:
# https://ui.adsabs.harvard.edu/abs/1973NASSP.330....../abstract
#
# .. _4:
#
# [4] L. Kauder, “Spacecraft Thermal Control Coatings References,” Goddard
# Space Flight Center, Greenbelt, MD 20771, NASA Technical Report
# NASA/TP-2005-212792, Dec. 2005. [Online]. Available:
# https://ntrs.nasa.gov/citations/20070014757



#
# This is an auto-generated file.  DO NOT EDIT!
#
# pylint: disable=line-too-long

from ansys.fluent.core.services.datamodel_se import (
    PyMenu,
    PyParameter,
    PyTextual,
    PyNumerical,
    PyDictionary,
    PyNamedObjectContainer,
    PyCommand,
    PyQuery
)


class Root(PyMenu):
    """
    Singleton Root.
    """
    def __init__(self, service, rules, path):
        self.GlobalSettings = self.__class__.GlobalSettings(service, rules, path + [("GlobalSettings", "")])
        self.JournalCommand = self.__class__.JournalCommand(service, rules, "JournalCommand", path)
        self.TWF_AssociateMesh = self.__class__.TWF_AssociateMesh(service, rules, "TWF_AssociateMesh", path)
        self.TWF_BasicMachineDescription = self.__class__.TWF_BasicMachineDescription(service, rules, "TWF_BasicMachineDescription", path)
        self.TWF_BladeRowAnalysisScope = self.__class__.TWF_BladeRowAnalysisScope(service, rules, "TWF_BladeRowAnalysisScope", path)
        self.TWF_CompleteWorkflowSetup = self.__class__.TWF_CompleteWorkflowSetup(service, rules, "TWF_CompleteWorkflowSetup", path)
        self.TWF_CreateCFDModel = self.__class__.TWF_CreateCFDModel(service, rules, "TWF_CreateCFDModel", path)
        self.TWF_ImportMesh = self.__class__.TWF_ImportMesh(service, rules, "TWF_ImportMesh", path)
        self.TWF_MapRegionInfo = self.__class__.TWF_MapRegionInfo(service, rules, "TWF_MapRegionInfo", path)
        self.TWF_ReportDefMonitors = self.__class__.TWF_ReportDefMonitors(service, rules, "TWF_ReportDefMonitors", path)
        self.TWF_TurboPhysics = self.__class__.TWF_TurboPhysics(service, rules, "TWF_TurboPhysics", path)
        self.TWF_TurboRegionsZones = self.__class__.TWF_TurboRegionsZones(service, rules, "TWF_TurboRegionsZones", path)
        self.TWF_TurboSurfaces = self.__class__.TWF_TurboSurfaces(service, rules, "TWF_TurboSurfaces", path)
        self.TWF_TurboTopology = self.__class__.TWF_TurboTopology(service, rules, "TWF_TurboTopology", path)
        super().__init__(service, rules, path)

    class GlobalSettings(PyMenu):
        """
        Singleton GlobalSettings.
        """
        def __init__(self, service, rules, path):
            self.EnableTurboMeshing = self.__class__.EnableTurboMeshing(service, rules, path + [("EnableTurboMeshing", "")])
            super().__init__(service, rules, path)

        class EnableTurboMeshing(PyParameter):
            """
            Parameter EnableTurboMeshing of value type bool.
            """
            pass

    class JournalCommand(PyCommand):
        """
        Command JournalCommand.

        Parameters
        ----------
        JournalString : str

        Returns
        -------
        bool
        """
        pass

    class TWF_AssociateMesh(PyCommand):
        """
        Command TWF_AssociateMesh.

        Parameters
        ----------
        AMChildName : str
        AMSelectComponentScope : str
        UseWireframe : bool
        RenameCellZones : str
        DefaultAMRowNumList : list[str]
        DefaultAMCellZonesList : list[str]
        AMRowNumList : list[str]
        OldAMCellZonesList : list[str]
        NewAMCellZonesList : list[str]

        Returns
        -------
        bool
        """
        pass

    class TWF_BasicMachineDescription(PyCommand):
        """
        Command TWF_BasicMachineDescription.

        Parameters
        ----------
        ComponentType : str
        ComponentName : str
        NumRows : int
        RowNumList : list[str]
        OldRowNameList : list[str]
        NewRowNameList : list[str]
        OldRowTypeList : list[str]
        NewRowTypeList : list[str]
        OldNumOfBladesList : list[str]
        NewNumOfBladesList : list[str]
        OldEnableTipGapList : list[str]
        NewEnableTipGapList : list[str]
        CombustorType : str

        Returns
        -------
        bool
        """
        pass

    class TWF_BladeRowAnalysisScope(PyCommand):
        """
        Command TWF_BladeRowAnalysisScope.

        Parameters
        ----------
        ASChildName : str
        ASSelectComponent : str
        ASRowNumList : list[str]
        OldASIncludeRowList : list[str]
        NewASIncludeRowList : list[str]

        Returns
        -------
        bool
        """
        pass

    class TWF_CompleteWorkflowSetup(PyCommand):
        """
        Command TWF_CompleteWorkflowSetup.


        Returns
        -------
        bool
        """
        pass

    class TWF_CreateCFDModel(PyCommand):
        """
        Command TWF_CreateCFDModel.

        Parameters
        ----------
        CFDMChildName : str
        CFDMSelectMeshAssociation : str
        AxisOfRotation : str
        DelayCFDModelCreation : bool
        RestrictToFactors : bool
        EstimateNumBlades : bool
        CFDMRowNumList : list[str]
        OldCFDMNumOfBladesList : list[str]
        NewCFDMNumOfBladesList : list[str]
        OldCFDMModelBladesList : list[str]
        NewCFDMModelBladesList : list[str]
        OldCFDMAngleOffset : list[str]
        NewCFDMAngleOffset : list[str]
        OldCFDMBladesPerSectorList : list[str]
        NewCFDMBladesPerSectorList : list[str]

        Returns
        -------
        bool
        """
        pass

    class TWF_ImportMesh(PyCommand):
        """
        Command TWF_ImportMesh.

        Parameters
        ----------
        AddChild : str
        MeshFilePath : str
        MeshFilePath_old : str
        MeshName : str
        CellZoneNames : list[str]
        ListItemLevels : list[str]
        ListItemTitles : list[str]
        ListOfCellZones : str

        Returns
        -------
        bool
        """
        pass

    class TWF_MapRegionInfo(PyCommand):
        """
        Command TWF_MapRegionInfo.

        Parameters
        ----------
        MRChildName : str
        MRSelectCellZone : str
        UseWireframe : bool
        DefaultMRRegionNameList : list[str]
        DefaultMRFaceZoneList : list[str]
        MRRegionNameList : list[str]
        OldMRFaceZoneList : list[str]
        NewMRFaceZoneList : list[str]

        Returns
        -------
        bool
        """
        pass

    class TWF_ReportDefMonitors(PyCommand):
        """
        Command TWF_ReportDefMonitors.

        Parameters
        ----------
        RDIsoSurfaceNumList : list[str]
        OldCreateContourList : list[str]
        NewCreateContourList : list[str]
        TurboContoursList : list[str]

        Returns
        -------
        bool
        """
        pass

    class TWF_TurboPhysics(PyCommand):
        """
        Command TWF_TurboPhysics.

        Parameters
        ----------
        States : dict[str, Any]

        Returns
        -------
        bool
        """
        pass

    class TWF_TurboRegionsZones(PyCommand):
        """
        Command TWF_TurboRegionsZones.

        Parameters
        ----------
        States : dict[str, Any]

        Returns
        -------
        bool
        """
        pass

    class TWF_TurboSurfaces(PyCommand):
        """
        Command TWF_TurboSurfaces.

        Parameters
        ----------
        NumIsoSurfaces : int
        IsoSurfaceNumList : list[str]
        OldIsoSurfaceNameList : list[str]
        NewIsoSurfaceNameList : list[str]
        OldIsoSurfaceValueList : list[str]
        NewIsoSurfaceValueList : list[str]
        SurfacesList : list[str]

        Returns
        -------
        bool
        """
        pass

    class TWF_TurboTopology(PyCommand):
        """
        Command TWF_TurboTopology.

        Parameters
        ----------
        TopologyName : str
        UseWireframe : bool
        DefaultTopologyNameList : list[str]
        DefaultTopologyZoneList : list[str]
        TopologyNameList : list[str]
        OldTopologyZoneList : list[str]
        NewTopologyZoneList : list[str]

        Returns
        -------
        bool
        """
        pass




# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.mark.fluent_version(">=25.1")
def test_systemcoupling_mixing_elbow_settings_apis(mixing_elbow_case_data_session):
    """Test System Coupling related settings APIs"""
    solver = mixing_elbow_case_data_session
    # enable the feature flag to be able to make the queries
    solver.scheme_eval.scheme_eval("(enable-feature 'sc/participant-info)")
    elbow_fluid = "elbow-fluid"
    region_names = solver.settings.setup.models.system_coupling.get_all_regions()
    assert elbow_fluid in region_names
    # elbow fluid must be a volume
    assert (
        solver.settings.setup.models.system_coupling.get_topology(
            region_name=elbow_fluid
        )
        == "Volume"
    )


def _test_systemcoupling_mixing_elbow_settings_common(mixing_elbow_case_data_session):
    solver = mixing_elbow_case_data_session
    # check participant type, analysis type, regions, and variables
    assert solver.system_coupling.participant_type == "FLUENT"
    assert solver.system_coupling.get_analysis_type() == "Steady"
    regions = solver.system_coupling.get_regions()
    variables = solver.system_coupling.get_variables()
    # [wall-inlet, wall-elbow, elbow-fluid, hot-inlet, cold-inlet, outlet]
    assert len(regions) >= 6
    # [force, temp, htc, hflow, nwt, hrate, e-cond, lorentz-force]
    assert len(variables) >= 8


@pytest.mark.fluent_version(">=25.1")
def test_systemcoupling_mixing_elbow_settings(mixing_elbow_case_data_session):
    """Very superficial test of System Coupling related settings."""
    _test_systemcoupling_mixing_elbow_settings_common(mixing_elbow_case_data_session)


@pytest.mark.fluent_version(">=24.1,<25.1")
def test_systemcoupling_mixing_elbow_settings_legacy(mixing_elbow_case_data_session):
    """Test legacy implementation of getting System Coupling related settings."""
    _test_systemcoupling_mixing_elbow_settings_common(mixing_elbow_case_data_session)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.mark.settings_only
@pytest.mark.fluent_version(">=24.1")
def test_change_create_mixture(mixing_elbow_settings_session):
    solver_session = mixing_elbow_settings_session

    # Test turning on species transport model
    species_mdl = solver_session.setup.models.species.model

    assert species_mdl.option() == "off"
    species_mdl.option = "species-transport"
    assert species_mdl.option() == "species-transport"

    # Test command names list
    materials = solver_session.setup.materials
    assert sorted(materials.mixture.command_names) == sorted(
        ["delete", "list", "list_properties", "make_a_copy", "rename"]
    )

    # Test change/creating a mixture with custom species from template
    custom_species_1 = materials.fluid.create("custom-species-1")
    custom_species_2 = materials.fluid.create("custom-species-2")
    solver_session.tui.define.materials.change_create(
        "mixture-template",  # Change/create mixture-template
        "custom-mixture",  # Rename to `custom-mixture`
        "yes",  # Change mixture species
        "2",  # Set number of species to 2
        custom_species_1.name(),  # Set species 1
        custom_species_2.name(),  # Set species 2
        "0",  # No surface species
        "0",  # No site species
        "no",  # Do not change density
        "no",  # Do not change specific heat
        "no",  # Do not change thermal conductivity
        "no",  # Do not change viscosity
        "no",  # Do not change mass diffusivity
        "no",  # Do not change speed of sound
        "yes",  # Overwrite mixture-template
    )
    assert "custom-mixture" in materials.mixture.keys()
    assert "mixture-template" not in materials.mixture.keys()

    # Test that mixture contains correct species
    mix_species = solver_session.setup.materials.mixture[
        "custom-mixture"
    ].species.volumetric_species.keys()
    assert "custom-species-1" in mix_species
    assert "custom-species-2" in mix_species

    # Test copying a mixture
    materials.mixture.make_a_copy(
        from_="custom-mixture",
        to="custom-mixture-copy",
    )
    assert "custom-mixture-copy" in materials.mixture.keys()

    # Test changing cellzone mixture
    elbow_zone = solver_session.setup.cell_zone_conditions.fluid["elbow-fluid"]
    assert elbow_zone.material() == "custom-mixture"
    elbow_zone.material = "custom-mixture-copy"
    assert elbow_zone.material() == "custom-mixture-copy"

    # Test deleting a mixture
    materials.mixture.delete(name_list=["custom-mixture"])
    assert "custom-mixture" not in materials.mixture.keys()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_exhaust_system_settings_api:

Fault-tolerant meshing workflow using settings objects
------------------------------------------------------
This example sets up and solves a three-dimensional turbulent fluid flow
in a manifold exhaust system, which is common in the automotive industry,
using the PyFluent Settings API. Predicting the flow field in the area of the
mixing region is important to designing the junction properly.

This example uses the guided workflow for fault-tolerant meshing because it
is appropriate for geometries that can have imperfections, such as gaps and
leakages.

**Workflow tasks**

The fault-tolerant meshing workflow guides you through these tasks:

- Import a CAD geometry and manage individual parts
- Generate a surface mesh
- Cap inlets and outlets
- Extract a fluid region
- Define leakages
- Extract edge features
- Set up size controls
- Generate a volume mesh

**Problem description**

In the manifold exhaust system, air flows through the three inlets
with a uniform velocity of 1 m/s. The air then exits through the outlet.
A small pipe is placed in the main portion of the manifold where edge
extraction is considered. The example also includes a known small leakage
to demonstrate the automatic leakage detection aspects of the meshing workflow.
"""

###############################################################################
# Example Setup
# -------------
# Before you can use the fault-tolerant meshing workflow, you must set up the
# example and initialize this workflow.
#
# Perform required imports
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Perform required imports, which includes downloading and importing
# the geometry file.

# sphinx_gallery_thumbnail_path = '_static/exhaust_system_settings.png'
import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples

import_file_name = examples.download_file(
    "exhaust_system.fmd", "pyfluent/exhaust_system"
)

###############################################################################
# Launch Fluent
# ~~~~~~~~~~~~~
# Launch Fluent as a service in meshing mode with double precision running on
# two processors and print Fluent version.

meshing = pyfluent.launch_fluent(
    precision="double",
    processor_count=2,
    mode="meshing",
)
print(meshing.get_fluent_version())

###############################################################################
# Initialize workflow
# ~~~~~~~~~~~~~~~~~~~
# Initialize the fault-tolerant meshing workflow.

meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")

###############################################################################
# Fault-folerant meshing workflow
# -------------------------------
# The fault-tolerant meshing workflow guides you through the many tasks that
# follow.
#
# Import CAD and manage parts
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Import the CAD geometry file (``exhaust_system.fmd``) and selectively manage some
# parts.

meshing.upload(import_file_name)
meshing.PartManagement.InputFileChanged(
    FilePath=import_file_name, IgnoreSolidNames=False, PartPerBody=False
)
meshing.PMFileManagement.FileManager.LoadFiles()
meshing.PartManagement.Node["Meshing Model"].Copy(
    Paths=[
        "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/main,1",
        "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/flow-pipe,1",
        "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/outpipe3,1",
        "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/object2,1",
        "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/object1,1",
    ]
)
meshing.PartManagement.ObjectSetting["DefaultObjectSetting"].OneZonePer.set_state(
    "part"
)
cad_import = meshing.workflow.TaskObject["Import CAD and Part Management"]
cad_import.Arguments.set_state(
    {
        "Context": 0,
        "CreateObjectPer": "Custom",
        "FMDFileName": import_file_name,
        "FileLoaded": "yes",
        "ObjectSetting": "DefaultObjectSetting",
        "Options": {
            "Line": False,
            "Solid": False,
            "Surface": False,
        },
    }
)
cad_import.Execute()

###############################################################################
# Describe geometry and flow
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# Describe the geometry and the flow characteristics.

describe_geom = meshing.workflow.TaskObject["Describe Geometry and Flow"]
describe_geom.Arguments.set_state(
    {
        "AddEnclosure": "No",
        "CloseCaps": "Yes",
        "FlowType": "Internal flow through the object",
    }
)
describe_geom.UpdateChildTasks(SetupTypeChanged=False)
describe_geom.Arguments.set_state(
    {
        "AddEnclosure": "No",
        "CloseCaps": "Yes",
        "DescribeGeometryAndFlowOptions": {
            "AdvancedOptions": True,
            "ExtractEdgeFeatures": "Yes",
        },
        "FlowType": "Internal flow through the object",
    }
)
describe_geom.UpdateChildTasks(SetupTypeChanged=False)
describe_geom.Execute()

###############################################################################
# Enclose openings
# ~~~~~~~~~~~~~~~~
# Enclose (cap) any openings in the geometry.

###############################################################################
# .. image:: /_static/exhaust_system_011.png
#   :width: 400pt
#   :align: center

###############################################################################
# .. image:: /_static/exhaust_system_012.png
#   :width: 400pt
#   :align: center
capping = meshing.workflow.TaskObject["Enclose Fluid Regions (Capping)"]
capping.Arguments.set_state(
    {
        "CreatePatchPreferences": {
            "ShowCreatePatchPreferences": False,
        },
        "PatchName": "inlet-1",
        "SelectionType": "zone",
        "ZoneSelectionList": ["inlet.1"],
    }
)
capping.Arguments.set_state(
    {
        "CreatePatchPreferences": {
            "ShowCreatePatchPreferences": False,
        },
        "PatchName": "inlet-1",
        "SelectionType": "zone",
        "ZoneLocation": [
            "1",
            "351.68205",
            "-361.34322",
            "-301.88668",
            "396.96205",
            "-332.84759",
            "-266.69751",
            "inlet.1",
        ],
        "ZoneSelectionList": ["inlet.1"],
    }
)
capping.AddChildToTask()

capping.InsertCompoundChildTask()
capping.Arguments.set_state({})
meshing.workflow.TaskObject["inlet-1"].Execute()
capping.Arguments.set_state(
    {
        "PatchName": "inlet-2",
        "SelectionType": "zone",
        "ZoneSelectionList": ["inlet.2"],
    }
)
capping.Arguments.set_state(
    {
        "PatchName": "inlet-2",
        "SelectionType": "zone",
        "ZoneLocation": [
            "1",
            "441.68205",
            "-361.34322",
            "-301.88668",
            "486.96205",
            "-332.84759",
            "-266.69751",
            "inlet.2",
        ],
        "ZoneSelectionList": ["inlet.2"],
    }
)
capping.AddChildToTask()

capping.InsertCompoundChildTask()
capping.Arguments.set_state({})
meshing.workflow.TaskObject["inlet-2"].Execute()
capping.Arguments.set_state(
    {
        "PatchName": "inlet-3",
        "SelectionType": "zone",
        "ZoneSelectionList": ["inlet"],
    }
)
capping.Arguments.set_state(
    {
        "PatchName": "inlet-3",
        "SelectionType": "zone",
        "ZoneLocation": [
            "1",
            "261.68205",
            "-361.34322",
            "-301.88668",
            "306.96205",
            "-332.84759",
            "-266.69751",
            "inlet",
        ],
        "ZoneSelectionList": ["inlet"],
    }
)
capping.AddChildToTask()

capping.InsertCompoundChildTask()
capping.Arguments.set_state({})
meshing.workflow.TaskObject["inlet-3"].Execute()
capping.Arguments.set_state(
    {
        "PatchName": "outlet-1",
        "SelectionType": "zone",
        "ZoneSelectionList": ["outlet"],
        "ZoneType": "pressure-outlet",
    }
)
capping.Arguments.set_state(
    {
        "PatchName": "outlet-1",
        "SelectionType": "zone",
        "ZoneLocation": [
            "1",
            "352.22702",
            "-197.8957",
            "84.102381",
            "394.41707",
            "-155.70565",
            "84.102381",
            "outlet",
        ],
        "ZoneSelectionList": ["outlet"],
        "ZoneType": "pressure-outlet",
    }
)
capping.AddChildToTask()

capping.InsertCompoundChildTask()
capping.Arguments.set_state({})
meshing.workflow.TaskObject["outlet-1"].Execute()

###############################################################################
# Extract edge features
# ~~~~~~~~~~~~~~~~~~~~~
# Extract edge features.
edge_features = meshing.workflow.TaskObject["Extract Edge Features"]
edge_features.Arguments.set_state(
    {
        "ExtractMethodType": "Intersection Loops",
        "ObjectSelectionList": ["flow_pipe", "main"],
    }
)
edge_features.AddChildToTask()

edge_features.InsertCompoundChildTask()
edge_group = meshing.workflow.TaskObject["edge-group-1"]
edge_group.Arguments.set_state(
    {
        "ExtractEdgesName": "edge-group-1",
        "ExtractMethodType": "Intersection Loops",
        "ObjectSelectionList": ["flow_pipe", "main"],
    }
)
edge_features.Arguments.set_state({})

edge_group.Execute()

###############################################################################
# Identify regions
# ~~~~~~~~~~~~~~~~
# Identify regions.
identify_regions = meshing.workflow.TaskObject["Identify Regions"]
identify_regions.Arguments.set_state(
    {
        "SelectionType": "zone",
        "X": 377.322045740589,
        "Y": -176.800676988458,
        "Z": -37.0764628583475,
        "ZoneSelectionList": ["main.1"],
    }
)
identify_regions.Arguments.set_state(
    {
        "SelectionType": "zone",
        "X": 377.322045740589,
        "Y": -176.800676988458,
        "Z": -37.0764628583475,
        "ZoneLocation": [
            "1",
            "213.32205",
            "-225.28068",
            "-158.25531",
            "541.32205",
            "-128.32068",
            "84.102381",
            "main.1",
        ],
        "ZoneSelectionList": ["main.1"],
    }
)
identify_regions.AddChildToTask()

identify_regions.InsertCompoundChildTask()
fluid_region_1 = meshing.workflow.TaskObject["fluid-region-1"]
fluid_region_1.Arguments.set_state(
    {
        "MaterialPointsName": "fluid-region-1",
        "SelectionType": "zone",
        "X": 377.322045740589,
        "Y": -176.800676988458,
        "Z": -37.0764628583475,
        "ZoneLocation": [
            "1",
            "213.32205",
            "-225.28068",
            "-158.25531",
            "541.32205",
            "-128.32068",
            "84.102381",
            "main.1",
        ],
        "ZoneSelectionList": ["main.1"],
    }
)
identify_regions.Arguments.set_state({})

fluid_region_1.Execute()
identify_regions.Arguments.set_state(
    {
        "MaterialPointsName": "void-region-1",
        "NewRegionType": "void",
        "ObjectSelectionList": ["inlet-1", "inlet-2", "inlet-3", "main"],
        "X": 374.722045740589,
        "Y": -278.9775145640143,
        "Z": -161.1700719416913,
    }
)
identify_regions.AddChildToTask()

identify_regions.InsertCompoundChildTask()

identify_regions.Arguments.set_state({})

meshing.workflow.TaskObject["void-region-1"].Execute()

###############################################################################
# Define thresholds for leakages
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Define thresholds for potential leakages.
leakage_threshold = meshing.workflow.TaskObject["Define Leakage Threshold"]
leakage_threshold.Arguments.set_state(
    {
        "AddChild": "yes",
        "FlipDirection": True,
        "PlaneDirection": "X",
        "RegionSelectionSingle": "void-region-1",
    }
)
leakage_threshold.AddChildToTask()

leakage_threshold.InsertCompoundChildTask()
leakage_1 = meshing.workflow.TaskObject["leakage-1"]
leakage_1.Arguments.set_state(
    {
        "AddChild": "yes",
        "FlipDirection": True,
        "LeakageName": "leakage-1",
        "PlaneDirection": "X",
        "RegionSelectionSingle": "void-region-1",
    }
)
leakage_threshold.Arguments.set_state(
    {
        "AddChild": "yes",
    }
)
leakage_1.Execute()

###############################################################################
# Review region settings
# ~~~~~~~~~~~~~~~~~~~~~~
# Review the region settings.
update_region = meshing.workflow.TaskObject["Update Region Settings"]
update_region.Arguments.set_state(
    {
        "AllRegionFilterCategories": ["2"] * 5 + ["1"] * 2,
        "AllRegionLeakageSizeList": ["none"] * 6 + ["6.4"],
        "AllRegionLinkedConstructionSurfaceList": ["n/a"] * 6 + ["no"],
        "AllRegionMeshMethodList": ["none"] * 6 + ["wrap"],
        "AllRegionNameList": [
            "main",
            "flow_pipe",
            "outpipe3",
            "object2",
            "object1",
            "void-region-1",
            "fluid-region-1",
        ],
        "AllRegionOversetComponenList": ["no"] * 7,
        "AllRegionSourceList": ["object"] * 5 + ["mpt"] * 2,
        "AllRegionTypeList": ["void"] * 6 + ["fluid"],
        "AllRegionVolumeFillList": ["none"] * 6 + ["tet"],
        "FilterCategory": "Identified Regions",
        "OldRegionLeakageSizeList": [""],
        "OldRegionMeshMethodList": ["wrap"],
        "OldRegionNameList": ["fluid-region-1"],
        "OldRegionOversetComponenList": ["no"],
        "OldRegionTypeList": ["fluid"],
        "OldRegionVolumeFillList": ["hexcore"],
        "RegionLeakageSizeList": [""],
        "RegionMeshMethodList": ["wrap"],
        "RegionNameList": ["fluid-region-1"],
        "RegionOversetComponenList": ["no"],
        "RegionTypeList": ["fluid"],
        "RegionVolumeFillList": ["tet"],
    }
)
update_region.Execute()


###############################################################################
# Set mesh control options
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Set mesh control options.

meshing.workflow.TaskObject["Choose Mesh Control Options"].Execute()

###############################################################################
# Generate surface mesh
# ~~~~~~~~~~~~~~~~~~~~~
# Generate the surface mesh.

###############################################################################
# .. image:: /_static/exhaust_system_013.png
#   :width: 500pt
#   :align: center

meshing.workflow.TaskObject["Generate the Surface Mesh"].Execute()

###############################################################################
# Confirm and update boundaries
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Confirm and update the boundaries.

meshing.workflow.TaskObject["Update Boundaries"].Execute()

###############################################################################
# Add boundary layers
# ~~~~~~~~~~~~~~~~~~~
# Add boundary layers.

meshing.workflow.TaskObject["Add Boundary Layers"].AddChildToTask()

meshing.workflow.TaskObject["Add Boundary Layers"].InsertCompoundChildTask()

meshing.workflow.TaskObject["aspect-ratio_1"].Arguments.set_state(
    {
        "BLControlName": "aspect-ratio_1",
    }
)
meshing.workflow.TaskObject["Add Boundary Layers"].Arguments.set_state({})

meshing.workflow.TaskObject["aspect-ratio_1"].Execute()

###############################################################################
# Generate volume mesh
# ~~~~~~~~~~~~~~~~~~~~
# Generate the volume mesh.

###############################################################################
# .. image:: /_static/exhaust_system_014.png
#   :width: 500pt
#   :align: center
volume_mesh_gen = meshing.workflow.TaskObject["Generate the Volume Mesh"]
volume_mesh_gen.Arguments.set_state(
    {
        "AllRegionNameList": [
            "main",
            "flow_pipe",
            "outpipe3",
            "object2",
            "object1",
            "void-region-1",
            "fluid-region-1",
        ],
        "AllRegionSizeList": ["11.33375"] * 7,
        "AllRegionVolumeFillList": ["none"] * 6 + ["tet"],
        "EnableParallel": True,
    }
)
volume_mesh_gen.Execute()

###############################################################################
# Check mesh
# ~~~~~~~~~~
# Check the mesh.

meshing.tui.mesh.check_mesh()

###############################################################################
# Solve and postprocess
# ---------------------
# Once you have completed the fault tolerate meshing workflow, you can solve and
# postprcess the results.
#
# Switch to solution mode
# ~~~~~~~~~~~~~~~~~~~~~~~
# Switch to the solution mode.

solver = meshing.switch_to_solver()

solver.mesh.check()

###############################################################################
# Select turbulence model
# ~~~~~~~~~~~~~~~~~~~~~~~
# Select the kw sst turbulence model.

viscous = solver.setup.models.viscous

viscous.model = "k-omega"
viscous.k_omega_model = "sst"

###############################################################################
# Set velocity and turbulence boundary conditions for first inlet
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Set the velocity and turbulence boundary conditions for the first inlet
# (``inlet-1``).

boundary_conditions = solver.setup.boundary_conditions

boundary_conditions.velocity_inlet["inlet-1"] = {
    "momentum": {
        "velocity_specification_method": "Magnitude, Normal to Boundary",
        "velocity": {
            "value": 1,
        },
    },
}

###############################################################################
# Set same boundary conditions for other velocity inlets
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Set the same boundary conditions for the other velocity inlets (``inlet_2``
# and ``inlet_3``).

boundary_conditions.copy(
    from_="inlet-1",
    to=["inlet-2", "inlet-3"],
)

###############################################################################
# Set boundary conditions at outlet
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Set the boundary conditions at the outlet (``outlet-1``).

boundary_conditions.pressure_outlet["outlet-1"].turbulence.turbulent_intensity = 0.05

###############################################################################
# Initialize flow field
# ~~~~~~~~~~~~~~~~~~~~~
# Initialize the flow field using hybrid initialization.

solver.solution.initialization.hybrid_initialize()

###############################################################################
# Start calculation
# ~~~~~~~~~~~~~~~~~
# Start the calculation by requesting 100 iterations.

###############################################################################
# .. image:: /_static/exhaust_system_015.png
#   :width: 500pt
#   :align: center

solver.solution.run_calculation.iterate(iter_count=100)

###############################################################################
# Write the case and data files
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

solver.file.write(
    file_type="case-data",
    file_name="exhaust_system.cas.h5",
)

###############################################################################
# Configure graphics picture export
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Since Fluent is being run without the GUI, we will need to export plots as
# picture files. Edit the picture settings to use a custom resolution so that
# the images are large enough.

graphics = solver.results.graphics
# use_window_resolution option not active inside containers or Ansys Lab environment
if graphics.picture.use_window_resolution.is_active():
    graphics.picture.use_window_resolution = False
graphics.picture.x_resolution = 1920
graphics.picture.y_resolution = 1440

###############################################################################
# Create path lines
# ~~~~~~~~~~~~~~~~~
# Create path lines highlighting the flow field, display it, then export the
# image for inspection.

###############################################################################
# .. image:: /_static/exhaust_system_016.png
#   :width: 500pt
#   :align: center

graphics.pathline["pathlines-1"] = {
    "field": "time",
    "accuracy_control": {
        "tolerance": 0.001,
    },
    "skip": 5,
    "release_from_surfaces": ["inlet-1", "inlet-2", "inlet-3"],
}
graphics.pathline["pathlines-1"].display()

graphics.views.restore_view(view_name="isometric")
graphics.views.auto_scale()
graphics.picture.save_picture(file_name="pathlines-1.png")

###############################################################################
# Create iso-surface
# ~~~~~~~~~~~~~~~~~~
# Create an iso-surface through the manifold geometry.

solver.results.surfaces.iso_surface["surf-x-coordinate"] = {
    "field": "x-coordinate",
    "zones": ["fluid-region-1"],
    "iso_values": [0.38],
}

###############################################################################
# Create contours of velocity magnitude
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Create contours of the velocity magnitude throughout the manifold
# along with the mesh. Display it and export the image for inspection.

###############################################################################
# .. image:: /_static/exhaust_system_017.png
#   :width: 500pt
#   :align: center

graphics.contour["contour-velocity"] = {
    "field": "velocity-magnitude",
    "surfaces_list": ["surf-x-coordinate"],
    "node_values": False,
    "range_option": {
        "option": "auto-range-on",
        "auto_range_on": {
            "global_range": False,
        },
    },
}
graphics.mesh["mesh-1"] = {
    "surfaces_list": "*",
}
graphics.contour["contour-velocity"].display()

graphics.views.restore_view(view_name="right")
graphics.views.auto_scale()
graphics.picture.save_picture(file_name="contour-velocity.png")

###############################################################################
# Create scene
# ~~~~~~~~~~~~
# Create a scene containing the mesh and the contours. Display it and export
# the image for inspection.

###############################################################################
# .. image:: /_static/exhaust_system_018.png
#   :width: 500pt
#   :align: center

solver.results.scene["scene-1"] = {}
scene1 = solver.results.scene["scene-1"]
scene1.graphics_objects.add(name="mesh-1")
scene1.graphics_objects["mesh-1"].transparency = 90
scene1.graphics_objects.add(name="contour-velocity")
scene1.display()

graphics.views.camera.position(xyz=[1.70, 1.14, 0.29])
graphics.views.camera.up_vector(xyz=[-0.66, 0.72, -0.20])
graphics.picture.save_picture(file_name="scene-1.png")

#########################################################################
# Close Fluent
# ~~~~~~~~~~~~
# Close Fluent.

solver.exit()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.api.fluent.v0.variant_pb2 import Variant
import ansys.fluent.core as pyfluent
from ansys.fluent.core.data_model_cache import DataModelCache, NameKey
from ansys.fluent.core.services.datamodel_se import _convert_value_to_variant


class Fake:
    def __init__(self, path):
        if isinstance(path, str):
            self.path = [
                comp.split(":") if ":" in comp else [comp, ""]
                for comp in path.split("/")
            ]
        else:
            self.path = path


def test_data_model_cache():
    cache = DataModelCache()
    cache.set_state("x", Fake([("A", ""), ("x", "")]), 42.0)
    assert 42.0 == cache.get_state("x", Fake([("A", ""), ("x", "")]))
    assert dict(x=42.0) == cache.get_state("x", Fake([("A", "")]))
    assert DataModelCache.Empty == cache.get_state("x", Fake([("B", "")]))
    assert DataModelCache.Empty == cache.get_state("y", Fake([]))


@pytest.mark.parametrize(
    "initial_cache,rules,state,deleted_paths,final_cache",
    [
        (
            {"r1": {"A": {"B": "ab", "C": {}, "D:D-1": {"__iname__": "D1"}}}},
            "r1",
            None,
            ["X/D:D1", "A/D:Y1", "A/D:D1"],
            {"r1": {"A": {"B": "ab", "C": {}}}},
        ),
        ({"r1": {}}, "r1", {"A": True}, [], {"r1": {"A": True}}),
        ({"r1": {}}, "r1", {"A": 5}, [], {"r1": {"A": 5}}),
        ({"r1": {}}, "r1", {"A": 3.0}, [], {"r1": {"A": 3.0}}),
        ({"r1": {}}, "r1", {"A": "ab"}, [], {"r1": {"A": "ab"}}),
        ({"r1": {}}, "r1", {"A": [False, True]}, [], {"r1": {"A": [False, True]}}),
        ({"r1": {}}, "r1", {"A": [5, 10]}, [], {"r1": {"A": [5, 10]}}),
        ({"r1": {}}, "r1", {"A": [3.0, 6.0]}, [], {"r1": {"A": [3.0, 6.0]}}),
        ({"r1": {}}, "r1", {"A": ["ab", "cd"]}, [], {"r1": {"A": ["ab", "cd"]}}),
        ({"r1": {"A": {}}}, "r1", {"A": {"B": 5}}, [], {"r1": {"A": {"B": 5}}}),
        ({"r1": {"A": 5}}, "r1", {"A": {}}, [], {"r1": {"A": {}}}),
        ({"r1": {"A": 5}}, "r1", {"A": None}, [], {"r1": {"A": None}}),
        (
            {"r1": {"A": {}}},
            "r1",
            {"A": {"B": {"C": 5}}},
            [],
            {"r1": {"A": {"B": {"C": 5}}}},
        ),
        (
            {"r1": {"A": {}}},
            "r1",
            {"A": {"B:B1": {"_name_": "B-1", "C": 5.0}}},
            [],
            {"r1": {"A": {"B:B-1": {"__iname__": "B1", "_name_": "B-1", "C": 5.0}}}},
        ),
        (
            {"r1": {"A": {"B:B-1": {"__iname__": "B1", "_name_": "B-1", "C": 5.0}}}},
            "r1",
            {"A": {"B:B1": {"C": 7.0}}},
            [],
            {"r1": {"A": {"B:B-1": {"__iname__": "B1", "_name_": "B-1", "C": 7.0}}}},
        ),
        (
            {"r1": {}},
            "r1",
            {"B:B1": {"_name_": "B-1", "C": 5.0}},
            [],
            {"r1": {"B:B-1": {"__iname__": "B1", "_name_": "B-1", "C": 5.0}}},
        ),
        (
            {"r1": {"B:B-1": {"__iname__": "B1", "_name_": "B-1", "C": 5.0}}},
            "r1",
            {"B:B1": {"C": 7.0}},
            [],
            {"r1": {"B:B-1": {"__iname__": "B1", "_name_": "B-1", "C": 7.0}}},
        ),
        (
            {"r1": {}},
            "r1",
            {"A:A1": {"_name_": "A-1", "B:B1": {"_name_": "B-1", "D": 7.0}, "C": 5.0}},
            [],
            {
                "r1": {
                    "A:A-1": {
                        "__iname__": "A1",
                        "_name_": "A-1",
                        "B:B-1": {"__iname__": "B1", "_name_": "B-1", "D": 7.0},
                        "C": 5.0,
                    }
                }
            },
        ),
        (
            {"r1": {"B:B-1": {"__iname__": "B1", "_name_": "B-1", "C": 5.0}}},
            "r1",
            {"B:B1": {"_name_": "B-2", "C": 5.0}},
            ["B:B1"],
            {"r1": {"B:B-2": {"__iname__": "B1", "_name_": "B-2", "C": 5.0}}},
        ),
    ],
)
def test_update_cache_display_names_as_keys(
    initial_cache, rules, state, deleted_paths, final_cache
):
    cache = DataModelCache()
    cache_rules = cache.rules_str_to_cache
    cache_rules.clear()
    cache_rules.update(initial_cache)
    var = Variant()
    _convert_value_to_variant(state, var)
    cache.update_cache(rules, var, deleted_paths)
    assert cache_rules == final_cache


@pytest.mark.parametrize(
    "initial_cache,rules,state,deleted_paths,final_cache",
    [
        (
            {"r1": {"A": {"B": "ab", "C": {}, "D:D1": {}}}},
            "r1",
            None,
            ["X/D:D1", "A/D:Y1", "A/D:D1"],
            {"r1": {"A": {"B": "ab", "C": {}}}},
        ),
        ({"r1": {}}, "r1", {"A": True}, [], {"r1": {"A": True}}),
        ({"r1": {}}, "r1", {"A": 5}, [], {"r1": {"A": 5}}),
        ({"r1": {}}, "r1", {"A": 3.0}, [], {"r1": {"A": 3.0}}),
        ({"r1": {}}, "r1", {"A": "ab"}, [], {"r1": {"A": "ab"}}),
        ({"r1": {}}, "r1", {"A": [False, True]}, [], {"r1": {"A": [False, True]}}),
        ({"r1": {}}, "r1", {"A": [5, 10]}, [], {"r1": {"A": [5, 10]}}),
        ({"r1": {}}, "r1", {"A": [3.0, 6.0]}, [], {"r1": {"A": [3.0, 6.0]}}),
        ({"r1": {}}, "r1", {"A": ["ab", "cd"]}, [], {"r1": {"A": ["ab", "cd"]}}),
        ({"r1": {"A": {}}}, "r1", {"A": {"B": 5}}, [], {"r1": {"A": {"B": 5}}}),
        (
            {"r1": {"A": {}}},
            "r1",
            {"A": {"B": {"C": 5}}},
            [],
            {"r1": {"A": {"B": {"C": 5}}}},
        ),
        (
            {"r1": {"A": {}}},
            "r1",
            {"A": {"B:B1": {"_name_": "B-1", "C": 5.0}}},
            [],
            {"r1": {"A": {"B:B1": {"_name_": "B-1", "C": 5.0}}}},
        ),
        (
            {"r1": {"A": {"B:B1": {"_name_": "B-1", "C": 5.0}}}},
            "r1",
            {"A": {"B:B1": {"C": 7.0}}},
            [],
            {"r1": {"A": {"B:B1": {"_name_": "B-1", "C": 7.0}}}},
        ),
        (
            {"r1": {}},
            "r1",
            {"B:B1": {"_name_": "B-1", "C": 5.0}},
            [],
            {"r1": {"B:B1": {"_name_": "B-1", "C": 5.0}}},
        ),
        (
            {"r1": {"B:B1": {"_name_": "B-1", "C": 5.0}}},
            "r1",
            {"B:B1": {"C": 7.0}},
            [],
            {"r1": {"B:B1": {"_name_": "B-1", "C": 7.0}}},
        ),
        (
            {"r1": {}},
            "r1",
            {"A:A1": {"_name_": "A-1", "B:B1": {"_name_": "B-1", "D": 7.0}, "C": 5.0}},
            [],
            {
                "r1": {
                    "A:A1": {
                        "_name_": "A-1",
                        "B:B1": {"_name_": "B-1", "D": 7.0},
                        "C": 5.0,
                    }
                }
            },
        ),
        (
            {"r1": {"B:B1": {"_name_": "B-1", "C": 5.0}}},
            "r1",
            {"B:B1": {"_name_": "B-2", "C": 5.0}},
            ["B:B1"],
            {"r1": {"B:B1": {"_name_": "B-2", "C": 5.0}}},
        ),
    ],
)
def test_update_cache_internal_names_as_keys(
    initial_cache, rules, state, deleted_paths, final_cache
):
    cache = DataModelCache()
    cache.set_config("r1", "name_key", NameKey.INTERNAL)
    cache_rules = cache.rules_str_to_cache
    cache_rules.clear()
    cache_rules.update(initial_cache)
    var = Variant()
    _convert_value_to_variant(state, var)
    cache.update_cache(rules, var, deleted_paths)
    assert cache_rules == final_cache


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_get_cached_values_in_command_arguments(new_meshing_session):
    wt = new_meshing_session.watertight()
    geo_import = new_meshing_session.workflow.TaskObject["Import Geometry"]
    geo_import.Arguments = dict(FileName="Bob")
    geo_import.Arguments = dict(FileName=None)
    assert "FileName" in wt.import_geometry.command_arguments()
    assert wt.import_geometry.command_arguments()["FileName"] is None


@pytest.fixture
def display_names_as_keys_in_cache():
    DataModelCache.use_display_name = True
    yield
    DataModelCache.use_display_name = False


def test_display_names_as_keys(
    display_names_as_keys_in_cache, watertight_workflow_session
):
    cache = watertight_workflow_session._datamodel_service_se.cache
    assert "TaskObject:Import Geometry" in cache.rules_str_to_cache["workflow"]
    assert "TaskObject:TaskObject1" not in cache.rules_str_to_cache["workflow"]


def test_internal_names_as_keys(watertight_workflow_session):
    cache = watertight_workflow_session._datamodel_service_se.cache
    assert "TaskObject:Import Geometry" not in cache.rules_str_to_cache["workflow"]
    assert "TaskObject:TaskObject1" in cache.rules_str_to_cache["workflow"]


@pytest.mark.parametrize(
    "rules_cache,name_key_in_config,path,name_key,state",
    [
        ({"A": {"B": {"C": 2}}}, NameKey.INTERNAL, "A/B", NameKey.INTERNAL, {"C": 2}),
        (
            {"A": {"B:B1": {"C:C1": {"_name_": "C-1"}, "_name_": "B-1"}}},
            NameKey.INTERNAL,
            "A/B:B-1",
            NameKey.DISPLAY,
            {"C:C-1": {"_name_": "C-1"}, "_name_": "B-1"},
        ),
        (
            {"A": {"B:B1": {"C:C1": {"_name_": "C-1"}, "_name_": "B-1"}}},
            NameKey.INTERNAL,
            "A/B:B1",
            NameKey.DISPLAY,
            {"C:C-1": {"_name_": "C-1"}, "_name_": "B-1"},
        ),
        (
            {"A": {"B:B-1": {"C:C-1": {"__iname__": "C1"}, "__iname__": "B1"}}},
            NameKey.DISPLAY,
            "A/B:B1",
            NameKey.INTERNAL,
            {"C:C1": {"__iname__": "C1"}, "__iname__": "B1"},
        ),
        (
            {"A": {"B:B1": {"C": 2, "_name_": "B-1"}}},
            NameKey.INTERNAL,
            "A/B:B-2",
            NameKey.DISPLAY,
            DataModelCache.Empty,
        ),
    ],
)
def test_cache_get_state(rules_cache, name_key_in_config, path, name_key, state):
    cache = DataModelCache()
    rules = "x"
    cache.set_config(rules, "name_key", name_key_in_config)
    cache_rules = cache.rules_str_to_cache
    cache_rules.clear()
    cache_rules[rules] = rules_cache
    assert state == cache.get_state(rules, Fake(path), name_key)


@pytest.mark.parametrize(
    "initial_cache,name_key_in_config,path,value,final_cache",
    [
        ({"A": 2}, NameKey.INTERNAL, "A/B", 2, {"A": {"B": 2}}),
        (
            {"A": 2},
            NameKey.INTERNAL,
            "A/B",
            {"C": {"D": 2}},
            {"A": {"B": {"C": {"D": 2}}}},
        ),
        (
            {"A": {"B": 2}},
            NameKey.INTERNAL,
            "A/B",
            {"C": {"D": 2}},
            {"A": {"B": {"C": {"D": 2}}}},
        ),
        (
            {"A": {"B": {"C": 2}}},
            NameKey.INTERNAL,
            "A/B",
            {"C": {"D": 2}},
            {"A": {"B": {"C": {"D": 2}}}},
        ),
        (
            {"A": {"B": {"C": {"D": 1}}}},
            NameKey.INTERNAL,
            "A/B",
            {"C": {"D": 2}},
            {"A": {"B": {"C": {"D": 2}}}},
        ),
        (
            {"A": {"B:B1": {"C:C1": {"_name_": "C-1"}, "_name_": "B-1"}}},
            NameKey.INTERNAL,
            "A/B:B-1",
            {"C:C-1": {"D": 2}},
            {"A": {"B:B1": {"C:C1": {"_name_": "C-1", "D": 2}, "_name_": "B-1"}}},
        ),
        (
            {"A": {"B:B1": {"C:C1": {"_name_": "C-1"}, "_name_": "B-1"}}},
            NameKey.INTERNAL,
            "A/B:B-1",
            {"C:C-1": {"D:D-1": {"__iname__": "D1"}}},
            {
                "A": {
                    "B:B1": {
                        "C:C1": {
                            "_name_": "C-1",
                            "D:D1": {"__iname__": "D1", "_name_": "D-1"},
                        },
                        "_name_": "B-1",
                    }
                }
            },
        ),
        (
            {"A": {"B:B1": {"C:C1": {"_name_": "C-1"}, "_name_": "B-1"}}},
            NameKey.INTERNAL,
            "A/B:B1",
            {"C:C1": {"D:D1": {"_name_": "D-1"}}},
            {
                "A": {
                    "B:B1": {
                        "C:C1": {
                            "_name_": "C-1",
                            "D:D1": {"_name_": "D-1", "__iname__": "D1"},
                        },
                        "_name_": "B-1",
                    }
                }
            },
        ),
        (
            {"A": {"B:B-1": {"C:C-1": {"__iname__": "C1"}, "__iname__": "B1"}}},
            NameKey.DISPLAY,
            "A/B:B1",
            {"C:C1": {"D": 2}},
            {"A": {"B:B-1": {"C:C-1": {"__iname__": "C1", "D": 2}, "__iname__": "B1"}}},
        ),
        (
            {"A": {"B:B-1": {"C:C-1": {"__iname__": "C1"}, "__iname__": "B1"}}},
            NameKey.DISPLAY,
            "A/B:B1",
            {"C:C1": {"D:D1": {"_name_": "D-1"}}},
            {
                "A": {
                    "B:B-1": {
                        "C:C-1": {
                            "__iname__": "C1",
                            "D:D-1": {"_name_": "D-1", "__iname__": "D1"},
                        },
                        "__iname__": "B1",
                    }
                }
            },
        ),
        (
            {"A": {"B:B-1": {"C:C-1": {"__iname__": "C1"}, "__iname__": "B1"}}},
            NameKey.DISPLAY,
            "A/B:B-1",
            {"C:C-1": {"D:D-1": {"__iname__": "D1"}}},
            {
                "A": {
                    "B:B-1": {
                        "C:C-1": {
                            "__iname__": "C1",
                            "D:D-1": {"__iname__": "D1", "_name_": "D-1"},
                        },
                        "__iname__": "B1",
                    }
                }
            },
        ),
    ],
)
def test_cache_set_state(
    initial_cache,
    name_key_in_config,
    path,
    value,
    final_cache,
):
    cache = DataModelCache()
    rules = "x"
    cache.set_config(rules, "name_key", name_key_in_config)
    cache_rules = cache.rules_str_to_cache
    cache_rules.clear()
    cache_rules[rules] = initial_cache
    cache.set_state(rules, Fake(path), value)
    assert final_cache == cache_rules[rules]


@pytest.mark.fluent_version(">=23.2")
def test_cache_per_session():
    with (
        pyfluent.launch_fluent(mode="meshing") as m1,
        pyfluent.launch_fluent(mode="meshing") as m2,
    ):
        assert m1.meshing.GlobalSettings.EnableComplexMeshing()
        assert m2.meshing.GlobalSettings.EnableComplexMeshing()
        _ = m1.watertight()
        assert not m1.meshing.GlobalSettings.EnableComplexMeshing()
        assert m2.meshing.GlobalSettings.EnableComplexMeshing()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Wrapper over the monitor gRPC service of Fluent."""

from google.protobuf.json_format import MessageToDict
import grpc

from ansys.api.fluent.v0 import monitor_pb2 as MonitorModule
from ansys.api.fluent.v0 import monitor_pb2_grpc as MonitorGrpcModule
from ansys.fluent.core.services.interceptors import (
    BatchInterceptor,
    ErrorStateInterceptor,
    TracingInterceptor,
)
from ansys.fluent.core.services.streaming import StreamingService


class MonitorsService(StreamingService):
    """Class wrapping the monitor gRPC service of Fluent."""

    def __init__(self, channel: grpc.Channel, metadata, fluent_error_state):
        """__init__ method of MonitorsService class."""
        intercept_channel = grpc.intercept_channel(
            channel,
            ErrorStateInterceptor(fluent_error_state),
            TracingInterceptor(),
            BatchInterceptor(),
        )
        self._stub = MonitorGrpcModule.MonitorStub(intercept_channel)
        self._metadata = metadata
        super().__init__(
            stub=self._stub,
            metadata=self._metadata,
        )

    def get_monitors_info(self) -> dict:
        """Get monitors information.

        Parameters
        ----------
        None

        Returns
        -------
        dict
            Dictionary containing the monitors information.
        """
        monitors_info = {}
        request = MonitorModule.GetMonitorsRequest()
        response = self._stub.GetMonitors(request, metadata=self._metadata)
        for monitor_set in response.monitorset:
            monitor_info = MessageToDict(monitor_set)
            monitors_info[monitor_set.name] = monitor_info
        return monitors_info



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Wrapper to settings gRPC service of Fluent."""

import collections.abc
from functools import wraps
from typing import Any

import grpc

from ansys.api.fluent.v0 import settings_pb2 as SettingsModule
from ansys.api.fluent.v0 import settings_pb2_grpc as SettingsGrpcModule
from ansys.fluent.core.services.interceptors import (
    BatchInterceptor,
    ErrorStateInterceptor,
    GrpcErrorInterceptor,
    TracingInterceptor,
)


class _SettingsServiceImpl:
    def __init__(
        self, channel: grpc.Channel, metadata: list[tuple[str, str]], fluent_error_state
    ) -> None:
        intercept_channel = grpc.intercept_channel(
            channel,
            GrpcErrorInterceptor(),
            ErrorStateInterceptor(fluent_error_state),
            TracingInterceptor(),
            BatchInterceptor(),
        )
        self.__stub = SettingsGrpcModule.SettingsStub(intercept_channel)
        self.__metadata = metadata

    def set_var(
        self, request: SettingsModule.SetVarRequest
    ) -> SettingsModule.SetVarResponse:
        """Set a variable."""
        return self.__stub.SetVar(request, metadata=self.__metadata)

    def get_var(
        self, request: SettingsModule.GetVarRequest
    ) -> SettingsModule.GetVarResponse:
        """Get a variable."""
        return self.__stub.GetVar(request, metadata=self.__metadata)

    def rename(
        self, request: SettingsModule.RenameRequest
    ) -> SettingsModule.RenameResponse:
        """Rename an object."""
        return self.__stub.Rename(request, metadata=self.__metadata)

    def create(
        self, request: SettingsModule.CreateRequest
    ) -> SettingsModule.CreateResponse:
        """Create an object."""
        return self.__stub.Create(request, metadata=self.__metadata)

    def delete(
        self, request: SettingsModule.DeleteRequest
    ) -> SettingsModule.DeleteResponse:
        """Delete an object."""
        return self.__stub.Delete(request, metadata=self.__metadata)

    def get_object_names(
        self, request: SettingsModule.GetObjectNamesRequest
    ) -> SettingsModule.GetObjectNamesResponse:
        """Get object names."""
        return self.__stub.GetObjectNames(request, metadata=self.__metadata)

    def get_list_size(
        self, request: SettingsModule.GetListSizeRequest
    ) -> SettingsModule.GetListSizeResponse:
        """Get list size."""
        return self.__stub.GetListSize(request, metadata=self.__metadata)

    def resize_list_object(
        self, request: SettingsModule.ResizeListObjectRequest
    ) -> SettingsModule.ResizeListObjectResponse:
        """Resize list object."""
        return self.__stub.ResizeListObject(request, metadata=self.__metadata)

    def get_static_info(
        self, request: SettingsModule.GetStaticInfoRequest
    ) -> SettingsModule.GetStaticInfoResponse:
        """Get static info."""
        return self.__stub.GetStaticInfo(request, metadata=self.__metadata)

    def execute_cmd(
        self, request: SettingsModule.ExecuteCommandRequest
    ) -> SettingsModule.ExecuteCommandResponse:
        """Execute the command."""
        return self.__stub.ExecuteCommand(request, metadata=self.__metadata)

    def execute_query(
        self, request: SettingsModule.ExecuteQueryRequest
    ) -> SettingsModule.ExecuteQueryResponse:
        """Execute the query."""
        return self.__stub.ExecuteQuery(request, metadata=self.__metadata)

    def get_attrs(
        self, request: SettingsModule.GetAttrsRequest
    ) -> SettingsModule.GetAttrsResponse:
        """Get attributes."""
        return self.__stub.GetAttrs(request, metadata=self.__metadata)


trace: bool = False
_indent: int = 0


def _trace(fn):
    @wraps(fn)
    def _fn(self, *args, **kwds):
        global _indent
        if trace:
            print(f"{' '*_indent}fn={fn.__name__}, args={args} {{")
            try:
                _indent += 1
                ret = fn(self, *args, **kwds)
            finally:
                _indent -= 1
            print(f"{' '*_indent}fn = {fn.__name__}, ret={ret} }}")
            return ret
        else:
            return fn(self, *args, **kwds)

    return _fn


def _get_request_instance_for_path(request_class, path: str) -> Any:
    request = request_class()
    request.path_info.path = path
    request.path_info.root = "fluent"
    return request


class SettingsService:
    """Service for accessing and modifying Fluent settings."""

    def __init__(
        self, channel, metadata, app_utilities, scheme_eval, fluent_error_state
    ) -> None:
        """__init__ method of SettingsService class."""
        self._service_impl = _SettingsServiceImpl(channel, metadata, fluent_error_state)
        self._app_utilities = app_utilities
        self._scheme_eval = scheme_eval

    @_trace
    def _set_state_from_value(self, state: SettingsModule.Value, value: Any):
        if value is None:
            return
        if isinstance(value, bool):
            state.boolean = value
        elif isinstance(value, int):
            state.integer = value
        elif isinstance(value, float):
            state.real = value
        elif isinstance(value, str):
            state.string = value
        elif isinstance(value, collections.abc.Mapping):
            for k, v in value.items():
                self._set_state_from_value(state.value_map.m[k], v)
        elif isinstance(value, collections.abc.Iterable):
            for v in value:
                self._set_state_from_value(state.value_list.lst.add(), v)
        else:  # fall back to string (for example, pathlib.Path)
            state.string = str(value)

    @_trace
    def _get_state_from_value(self, state: SettingsModule.Value) -> Any:
        t = state.WhichOneof("value")
        if t == "boolean":
            return state.boolean
        elif t == "integer":
            return state.integer
        elif t == "real":
            return state.real
        elif t == "string":
            return state.string
        elif t == "value_list":
            return [self._get_state_from_value(v) for v in state.value_list.lst]
        elif t == "value_map":
            return {
                k: self._get_state_from_value(v)
                for k, v in sorted(state.value_map.m.items())
            }
        else:
            return None

    @_trace
    def set_var(self, path: str, value: Any) -> None:
        """Set the value for the given path."""
        request = _get_request_instance_for_path(SettingsModule.SetVarRequest, path)
        self._set_state_from_value(request.value, value)
        self._service_impl.set_var(request)

    @_trace
    def get_var(self, path: str) -> Any:
        """Get the value for the given path."""
        request = _get_request_instance_for_path(SettingsModule.GetVarRequest, path)
        response = self._service_impl.get_var(request)
        return self._get_state_from_value(response.value)

    @_trace
    def rename(self, path: str, new: str, old: str) -> None:
        """Rename the object at the given path."""
        request = _get_request_instance_for_path(SettingsModule.RenameRequest, path)
        request.old_name = old
        request.new_name = new

        self._service_impl.rename(request)

    @_trace
    def create(self, path: str, name: str) -> None:
        """Create a named object child for the given path."""
        request = _get_request_instance_for_path(SettingsModule.CreateRequest, path)
        request.name = name

        self._service_impl.create(request)

    @_trace
    def delete(self, path: str, name: str) -> None:
        """Delete the object with the given name at the given path."""
        request = _get_request_instance_for_path(SettingsModule.DeleteRequest, path)
        request.name = name

        self._service_impl.delete(request)

    @_trace
    def get_object_names(self, path: str) -> list[str]:
        """Get a list of named objects."""
        request = _get_request_instance_for_path(
            SettingsModule.GetObjectNamesRequest, path
        )
        return self._service_impl.get_object_names(request).names

    @_trace
    def get_list_size(self, path: str) -> int:
        """Get the number of elements in a list object."""
        request = _get_request_instance_for_path(
            SettingsModule.GetListSizeRequest, path
        )
        return self._service_impl.get_list_size(request).size

    @_trace
    def resize_list_object(self, path: str, size: int) -> None:
        """Resize a list object."""
        request = _get_request_instance_for_path(
            SettingsModule.ResizeListObjectRequest, path
        )
        request.size = size
        return self._service_impl.resize_list_object(request)

    @_trace
    def _extract_static_info(self, info: SettingsModule.StaticInfo) -> dict[str, Any]:
        ret = {}
        ret["type"] = info.type
        for key, value in sorted(info.attrs.items()):
            ret[key] = self._get_state_from_value(value)
        if info.has_allowed_values:
            ret["has-allowed-values"] = info.has_allowed_values
        if info.children:
            ret["children"] = {
                child.name: self._extract_static_info(child.value)
                for child in info.children
            }
        if info.commands:
            ret["commands"] = {
                child.name: self._extract_static_info(child.value)
                for child in info.commands
            }
        if hasattr(info, "queries") and info.queries:
            ret["queries"] = {
                child.name: self._extract_static_info(child.value)
                for child in info.queries
            }
        if info.arguments:
            ret["arguments"] = {
                child.name: self._extract_static_info(child.value)
                for child in info.arguments
            }
        if info.HasField("object_type"):
            ret["object-type"] = self._extract_static_info(info.object_type)
        if info.help:
            ret["help"] = info.help
        try:
            if info.include_child_named_objects:
                ret["include_child_named_objects"] = info.include_child_named_objects
        except AttributeError:
            pass

        try:
            if info.list_size:
                ret["list_size"] = info.list_size
        except AttributeError:
            pass

        try:
            if info.user_creatable:
                ret["user_creatable"] = info.user_creatable
        except AttributeError:
            ret["user_creatable"] = True

        return ret

    @_trace
    def get_static_info(self) -> dict[str, Any]:
        """Get static-info for settings.

        Raises
        ------
        RuntimeError
            If type is empty.
        """
        request = SettingsModule.GetStaticInfoRequest()
        request.root = "fluent"
        response = self._service_impl.get_static_info(request)
        # The RPC calls no longer raise an exception. Force an exception if
        # type is empty
        if not response.info.type:
            raise RuntimeError
        return self._extract_static_info(response.info)

    @_trace
    def execute_cmd(self, path: str, command: str, **kwds) -> Any:
        """Execute a given command with the provided keyword arguments."""
        request = _get_request_instance_for_path(
            SettingsModule.ExecuteCommandRequest, path
        )
        request.command = command
        self._set_state_from_value(request.args, kwds)

        response = self._service_impl.execute_cmd(request)
        return self._get_state_from_value(response.reply)

    @_trace
    def execute_query(self, path: str, query: str, **kwds) -> Any:
        """Execute a given query with the provided keyword arguments."""
        request = _get_request_instance_for_path(
            SettingsModule.ExecuteQueryRequest, path
        )
        request.query = query
        self._set_state_from_value(request.args, kwds)

        response = self._service_impl.execute_query(request)
        return self._get_state_from_value(response.reply)

    @_trace
    def _parse_attrs(self, response: SettingsModule.GetAttrsResponse) -> dict[str, Any]:
        ret = {}
        ret["attrs"] = self._get_state_from_value(response.values)
        if response.group_children:
            ret["group_children"] = {
                child.name: self._parse_attrs(child.value)
                for child in response.group_children
            }
        return ret

    @_trace
    def get_attrs(self, path: str, attrs: list[str], recursive: bool = False) -> Any:
        """Return values of given attributes."""
        request = _get_request_instance_for_path(SettingsModule.GetAttrsRequest, path)
        request.attrs[:] = attrs
        request.recursive = recursive

        response = self._service_impl.get_attrs(request)
        if recursive:
            return self._parse_attrs(response)
        return self._get_state_from_value(response.values)

    @_trace
    def has_wildcard(self, name: str) -> bool:
        """Checks whether a name has a wildcard pattern."""
        return self._scheme_eval.is_defined(
            "has-fnmatch-wild-card?"
        ) and self._app_utilities.is_wildcard(name)

    @_trace
    def is_interactive_mode(self) -> bool:
        """Checks whether commands can be executed interactively."""
        return False



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_parametric_static_mixer_1:

Parametric study workflow
-------------------------
This example shows how you can use the parametric study workflow to analyze a
static mixer.
"""

from pathlib import Path

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples

# sphinx_gallery_thumbnail_path = '_static/DP_table.png'
############################################################################
# Perform required imports
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Perform the required imports.


############################################################################
# Launch Fluent
# ~~~~~~~~~~~~~
# Launch Fluent in 3D and double precision and print Fluent version.

solver_session = pyfluent.launch_fluent(
    precision="double", processor_count=2, mode="solver"
)
print(solver_session.get_fluent_version())

############################################################################
# Download and read files
# ~~~~~~~~~~~~~~~~~~~~~~~
# Download the files for this example and read the case for the static mixer.

import_filename = examples.download_file(
    "Static_Mixer_main.cas.h5",
    "pyfluent/static_mixer",
    return_without_path=False,
)

solver_session.settings.file.read_case(file_name=import_filename)

############################################################################
# Set iterations
# ~~~~~~~~~~~~~~
# Set the number of iterations to 100.

solver_session.settings.solution.run_calculation.iter_count = 100

############################################################################
# Create input parameters
# ~~~~~~~~~~~~~~~~~~~~~~~
# Enable parameter creation in the TUI and then create input parameters for
# the velocity and temperatures of inlets 1 and 2:
# Parameter values:
# Inlet1: velocity (inlet1_vel) 0.5 m/s and temperature (inlet1_temp) at 300 K
# Inlet2: velocity (inlet2_vel) 0.5 m/s and temperature (inlet2_temp) at 350 K

solver_session.settings.parameters.enable_in_tui = True

solver_session.settings.setup.named_expressions["inlet1_vel"] = {
    "output_parameter": False,
    "input_parameter": True,
    "unit": "velocity",
    "parametername": "inlet1_vel",
    "parameterid": "real-1",
    "description": "",
    "definition": "1 [m/s]",
    "name": "inlet1_vel",
}

solver_session.settings.setup.boundary_conditions.velocity_inlet["inlet1"] = {
    "momentum": {"velocity": {"value": "inlet1_vel"}}
}

solver_session.settings.setup.named_expressions["inlet1_temp"] = {
    "output_parameter": False,
    "input_parameter": True,
    "unit": "temperature",
    "parametername": "inlet1_temp",
    "parameterid": "real-2",
    "description": "",
    "definition": "300 [K]",
    "name": "inlet1_temp",
}

solver_session.settings.setup.boundary_conditions.velocity_inlet["inlet1"] = {
    "thermal": {"temperature": {"value": "inlet1_temp"}}
}

solver_session.settings.setup.named_expressions["inlet2_vel"] = {
    "output_parameter": False,
    "input_parameter": True,
    "unit": "velocity",
    "parametername": "inlet2_vel",
    "parameterid": "real-3",
    "description": "",
    "definition": "1 [m/s]",
    "name": "inlet2_vel",
}

solver_session.settings.setup.boundary_conditions.velocity_inlet["inlet2"] = {
    "momentum": {"velocity": {"value": "inlet2_vel"}}
}

solver_session.settings.setup.named_expressions["inlet2_temp"] = {
    "output_parameter": False,
    "input_parameter": True,
    "unit": "temperature",
    "parametername": "inlet2_temp",
    "parameterid": "real-4",
    "description": "",
    "definition": "350 [K]",
    "name": "inlet2_temp",
}

solver_session.settings.setup.boundary_conditions.velocity_inlet["inlet2"] = {
    "thermal": {"temperature": {"value": "inlet2_temp"}}
}

###########################################################################
# Create output parameters
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Create output parameters named ``outlet-temp-avg`` and ``outlet-vel-avg``
# using report definitions.

solver_session.settings.solution.report_definitions.surface["outlet-temp-avg"] = {}
solver_session.settings.solution.report_definitions.surface[
    "outlet-temp-avg"
].report_type = "surface-areaavg"
solver_session.settings.solution.report_definitions.surface["outlet-temp-avg"].field = (
    "temperature"
)
solver_session.settings.solution.report_definitions.surface[
    "outlet-temp-avg"
].surface_names = ["outlet"]

solver_session.settings.solution.report_definitions.surface["outlet-vel-avg"] = {}
solver_session.settings.solution.report_definitions.surface[
    "outlet-vel-avg"
].report_type = "surface-areaavg"
solver_session.settings.solution.report_definitions.surface["outlet-vel-avg"].field = (
    "velocity-magnitude"
)
solver_session.settings.solution.report_definitions.surface[
    "outlet-vel-avg"
].surface_names = ["outlet"]

solver_session.settings.parameters.output_parameters.report_definitions[
    "parameter-5"
] = {}
solver_session.settings.parameters.output_parameters.report_definitions[
    "parameter-5"
] = {"report_definition": "outlet-temp-avg"}

solver_session.settings.parameters.output_parameters.report_definitions[
    "parameter-6"
] = {}
solver_session.settings.parameters.output_parameters.report_definitions[
    "parameter-6"
] = {"report_definition": "outlet-vel-avg"}

###########################################################################
# Enable convergence condition check
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Enable the convergence condition check.

solver_session.settings.solution.monitor.residual.options.criterion_type = "absolute"

###########################################################################
# Write case
# ~~~~~~~~~~
# Write the case with all settings in place.

case_path = str(Path(pyfluent.EXAMPLES_PATH) / "Static_Mixer_Parameters.cas.h5")
solver_session.settings.file.write_case(file_name=case_path)

###########################################################################
# Initialize parametric study
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Initialize a parametric design point study from a Fluent session.

solver_session.settings.parametric_studies.initialize(project_filename="project_1")

###############################################################################
# .. image:: /_static/DP_table_011.png
#   :width: 500pt
#   :align: center

###########################################################################
# Access and modify input parameters
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Access and modify the input parameters of the base design point.

solver_session.settings.parametric_studies["Static_Mixer_main-Solve"].design_points[
    "Base DP"
].input_parameters = {
    "inlet1_temp": 300,
    "inlet2_temp": 350,
    "inlet2_vel": 1,
    "inlet1_vel": 0.5,
}

###########################################################################
# Update current design point
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Update the current design point.

solver_session.settings.parametric_studies[
    "Static_Mixer_main-Solve"
].design_points.update_current()

###########################################################################
# Add design point
# ~~~~~~~~~~~~~~~~
# Add a design point.

solver_session.settings.parametric_studies[
    "Static_Mixer_main-Solve"
].design_points.create(write_data=False, capture_simulation_report_data=True)

solver_session.settings.parametric_studies["Static_Mixer_main-Solve"].design_points[
    "DP1"
].input_parameters = {
    "inlet1_vel": 1,
    "inlet2_vel": 1,
    "inlet1_temp": 500,
    "inlet2_temp": 350,
}

##########################################################################
# Duplicate design point
# ~~~~~~~~~~~~~~~~~~~~~~~
# Duplicate design point 1 to create design point 2.

solver_session.settings.parametric_studies[
    "Static_Mixer_main-Solve"
].design_points.duplicate(design_point="DP1")

#########################################################################
# Update all design points
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Update all design points.

solver_session.settings.parametric_studies[
    "Static_Mixer_main-Solve"
].design_points.update_all()

###############################################################################
# .. image:: /_static/DP_table_012.png
#   :width: 500pt
#   :align: center

###############################################################################
# Export design point table
# ~~~~~~~~~~~~~~~~~~~~~~~~~
# Export the design point table to a CSV file.

design_point_table = str(
    Path(pyfluent.EXAMPLES_PATH) / "design_point_table_study_1.csv"
)

solver_session.settings.parametric_studies.export_design_table(
    filepath=design_point_table
)

##########################################################################
# Delete design point
# ~~~~~~~~~~~~~~~~~~~
# Delete design point 1.

solver_session.settings.parametric_studies[
    "Static_Mixer_main-Solve"
].design_points.delete_design_points(design_points=["DP1"])

##########################################################################
# Create parametric study
# ~~~~~~~~~~~~~~~~~~~~~~~
# Create another parametric study by duplicating the current one.

solver_session.settings.parametric_studies.duplicate(copy_design_points=True)

#########################################################################
# Rename new parametric study
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Rename the newly created parametric study.

solver_session.settings.parametric_studies.rename(
    new="New Study", old="Static_Mixer_main-1-Solve"
)

#########################################################################
# Delete old parametric study
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Delete the old parametric study.

solver_session.settings.parametric_studies.delete(name_list="Static_Mixer_main-Solve")

#########################################################################
# Save parametric project and close Fluent
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Save the parametric project and close Fluent.

project_filepath = str(Path(pyfluent.EXAMPLES_PATH) / "static_mixer_study_save.flprj")

solver_session.settings.file.parametric_project.save_as(
    project_filename=project_filepath
)

solver_session.exit()

#########################################################################
# Launch Fluent and read saved project
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Launch Fluent once again and read the previously saved project.

solver_session = pyfluent.launch_fluent(
    precision="double", processor_count=2, mode="solver"
)

project_filepath_read = str(
    Path(pyfluent.EXAMPLES_PATH) / "static_mixer_study_save.flprj"
)

solver_session.settings.file.parametric_project.open(
    project_filename=project_filepath_read, load_case=True
)

#########################################################################
# Save current project
# ~~~~~~~~~~~~~~~~~~~~
# Save the current project.

solver_session.settings.file.parametric_project.save()

#########################################################################
# Save current project as a different project
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Save the current project as a different project.

project_filepath_save_as = str(
    Path(pyfluent.EXAMPLES_PATH) / "static_mixer_study_save_as.flprj"
)

solver_session.settings.file.parametric_project.save_as(
    project_filename=project_filepath_save_as
)

#########################################################################
# Export current project
# ~~~~~~~~~~~~~~~~~~~~~~
# Export the current project.

project_filepath_export = str(
    Path(pyfluent.EXAMPLES_PATH) / "static_mixer_study_export.flprj"
)

solver_session.settings.file.parametric_project.save_as_copy(
    project_filename=project_filepath_export, convert_to_managed=False
)

#########################################################################
# Archive current project
# ~~~~~~~~~~~~~~~~~~~~~~~
# Archive the current project.

project_filepath_archive = str(
    Path(pyfluent.EXAMPLES_PATH) / "static_mixer_study_archive.flprj"
)

solver_session.settings.file.parametric_project.archive(
    archive_name=project_filepath_archive
)

#########################################################################
# Close Fluent
# ~~~~~~~~~~~~
# Close Fluent.

solver_session.exit()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ansys.fluent.core import examples
from ansys.fluent.core.filereader.case_file import CaseFile
from ansys.fluent.core.filereader.data_file import DataFile


def test_data_reader_for_single_phase():
    case_file_name = examples.download_file(
        "elbow1.cas.h5", "pyfluent/file_session", return_without_path=False
    )

    data_file_name = examples.download_file(
        "elbow1.dat.h5", "pyfluent/file_session", return_without_path=False
    )

    reader = DataFile(
        data_file_name=data_file_name,
        case_file_handle=CaseFile(case_file_name=case_file_name),
    )

    assert reader.case_file == "elbow1.cas.h5"

    assert len(reader.variables()) == 80

    assert reader.get_phases() == ["phase-1"]

    assert len(reader.get_face_variables("phase-1")) == 30

    assert len(reader.get_cell_variables("phase-1")) == 14

    assert reader.get_cell_variables("phase-1") == [
        "SV_BF_V",
        "SV_D",
        "SV_DENSITY",
        "SV_H",
        "SV_K",
        "SV_LORENTZ_FORCE",
        "SV_MU_LAM",
        "SV_MU_T",
        "SV_P",
        "SV_T",
        "SV_U",
        "SV_V",
        "SV_W",
        "",
    ]

    assert len(reader.get_face_scalar_field_data("phase-1", "SV_DENSITY", 3)) == 3630

    assert len(reader.get_face_vector_field_data("phase-1", 3)) == 10890


def test_data_reader_for_multi_phase():
    case_file_name = examples.download_file(
        "mixing_elbow_mul_ph.cas.h5",
        "pyfluent/file_session",
        return_without_path=False,
    )
    data_file_name = examples.download_file(
        "mixing_elbow_mul_ph.dat.h5",
        "pyfluent/file_session",
        return_without_path=False,
    )

    reader = DataFile(
        data_file_name=data_file_name,
        case_file_handle=CaseFile(case_file_name=case_file_name),
    )  # Instantiate a DataFile class

    assert reader.case_file == "mixing_elbow_mul_ph.cas.h5"

    assert len(reader.variables()) == 80

    assert reader.get_phases() == [
        "phase-1",
        "phase-2",
        "phase-3",
        "phase-4",
    ]

    assert len(reader.get_face_variables("phase-1")) == 23

    assert len(reader.get_face_variables("phase-3")) == 13

    assert len(reader.get_cell_variables("phase-2")) == 14

    assert reader.get_cell_variables("phase-2") == [
        "SV_BF_V",
        "SV_DENSITY",
        "SV_DENSITY_M1",
        "SV_MU_LAM",
        "SV_MU_T",
        "SV_U",
        "SV_U_M1",
        "SV_V",
        "SV_VOF",
        "SV_VOF_M1",
        "SV_V_M1",
        "SV_W",
        "SV_W_M1",
        "",
    ]

    assert len(reader.get_face_scalar_field_data("phase-1", "SV_DENSITY", 33)) == 268



"""Provides a module for generating Fluent Datamodel and TUI RST files."""

import fnmatch
import os
import pathlib
from pathlib import Path
import re

_THIS_DIRNAME = os.path.dirname(__file__)


def _get_attribute_classes(menu: type):
    """Get attribute classes.

    Parameters
    ----------
    menu: type
        TUIMenu, TUIMethod

    Returns
    -------
    attribute_classes: list
        Attributes of ``menu``.
    """
    attribute_classes = []
    attributes_dict = dict(vars(menu))
    for attr_name, attr_value in attributes_dict.items():
        if not attr_name.startswith("__"):
            attribute_classes.append(attributes_dict[attr_name])
    return attribute_classes


def _get_attribute_classes_with_and_without_members(menu: type):
    """Get classes with and without sub members as attributes.

    Parameters
    ----------
    menu: type
        TUIMenu, TUIMethod

    Returns
    -------
    with_members: list
        Attributes of ``menu`` with sub ``menu``.
    without_members: list
        Attributes of ``menu`` without sub ``menu``.
    """
    with_members = []
    without_members = []
    menu_members = _get_attribute_classes(menu)
    for menu_member in menu_members:
        if _get_attribute_classes(menu_member):
            with_members.append(menu_member)
        else:
            without_members.append(menu_member)
    return with_members, without_members


def _add_with_without_members(with_member: type, all_menus: list):
    """Add both with and without members to ``all_menus`` list.

    Parameters
    ----------
    with_member: type
        Attribute of ``menu`` with sub ``menu``.
    all_menus: list
        ``all_menus`` list.

    Returns
    -------
    with_members: list
        Attributes of ``menu`` with sub ``menu``.
    all_menus: list
        ``all_menus`` list.
    """
    with_members, without_members = _get_attribute_classes_with_and_without_members(
        with_member
    )
    all_menus.append(
        dict(
            name=with_member, with_members=with_members, without_members=without_members
        )
    )
    return with_members, all_menus


def _get_attribute_classes_recursively(with_member: type, all_menus: list):
    """Generate all attribute classes with and without members recursively.

    Parameters
    ----------
    with_member: type
        Attribute of ``menu`` with sub ``menu``.
    all_menus: list
        ``all_menus`` list.
    """
    with_members, all_menus = _add_with_without_members(with_member, all_menus)
    for with_member in with_members:
        _get_attribute_classes_recursively(with_member, all_menus)


def _process_datamodel_path(full_name: str):
    """Get path for datamodel files.

    Parameters
    ----------
    full_name: str
        Full name of class.

    Returns:
    path: str
        Path of datamodel class.
    """
    path_string = re.findall("core.*", full_name)
    path = path_string[0].replace("core.generated.", "")
    path = re.sub("[0-9]", "", path)
    path = path.replace("Root." if "Root." in path else "Root", "")
    path = path.replace("datamodel_.", "")
    path = path.rstrip(".")
    path = path.replace(".", "/")
    return path


def _process_tui_path(full_name: str):
    """Get path for tui files.

    Parameters
    ----------
    full_name: str
        Full name of class.

    Returns:
    path: str
        Path of tui class.
    """
    path_string = re.findall("main_menu.*", full_name)
    path = path_string[0].replace("main_menu", "")
    path = path.lstrip(".")
    path = path.replace(".", "/")
    return path


def _get_menu_name_path(menu: type, is_datamodel: bool):
    """Get menu name and path to generate RST file and folder respectively.

    Parameters
    ----------
    menu: type
        TUIMenu, TUIMethod
    is_datamodel: bool
        Whether to select datamodel menu.

    Returns
    -------
    full_name: str
        Full name of ``TUIMenu`` or ``TUIMethod``.
    full_path: str
        Full path of ``TUIMenu`` or ``TUIMethod``.
    """
    name_string = re.findall("ansys.*", str(menu))
    full_name = name_string[0][0:-2]
    if is_datamodel:
        full_path = _process_datamodel_path(full_name)
    else:
        full_path = _process_tui_path(full_name)
    return full_name, full_path


def _get_docdir(mode: str, path: str | None = None, is_datamodel: bool | None = None):
    """Get tui doc directory to generate all RST files.

    Parameters
    ----------
    mode: str
        Fluent session mode either ``meshing`` or ``solver``.
    path: str
        Full path of ``TUIMenu`` or ``TUIMethod``.

    Returns
    -------
    doc_path: str
        TUI doc directory.
    """
    doc_path = pathlib.Path(os.path.normpath(os.path.join(_THIS_DIRNAME, "..")))
    if is_datamodel:
        return doc_path / f"doc/source/api/{mode}/datamodel/{path}"
    else:
        return doc_path / f"doc/source/api/{mode}/tui/{path}"


def _get_path(mode: str, is_datamodel: bool | None = None):
    """Get datamodel_* or tui_*.py file path.

    Parameters
    ----------
    mode: str
        Fluent session mode either ``meshing`` or ``solver``.
    is_datamodel: bool
        Whether to get datamodel path.

    Returns
    -------
        Datamodel or TUI path.
    """
    if is_datamodel:
        return os.path.normpath(
            os.path.join(
                _THIS_DIRNAME,
                "..",
                "src",
                "ansys",
                "fluent",
                "core",
                "generated",
            )
        )
    else:
        return os.path.normpath(
            os.path.join(
                _THIS_DIRNAME,
                "..",
                "src",
                "ansys",
                "fluent",
                "core",
                "generated",
                f"{mode}",
            )
        )


def _get_file_or_folder(mode: str, is_datamodel: bool):
    """Get datamodel_* folder or tui_*.py file name.

    Parameters
    ----------
    mode: str
        Fluent session mode. Options are ``meshing`` and ``solver``.
    is_datamodel: bool
        Whether to get datamodel folder.

    Returns
    -------
        Datamodel or TUI file name.
    """
    for file in os.listdir(_get_path(mode, is_datamodel)):
        if is_datamodel and fnmatch.fnmatch(file, "datamodel_*"):
            return pathlib.Path(file).stem
        if not is_datamodel and fnmatch.fnmatch(file, "tui_*.py"):
            return pathlib.Path(file).stem


def _get_sorted_members(members: list):
    """Sort members alphabetically.

    Parameters
    ----------
    members: list
        Attributes of ``TUIMenu`` or ``TUIMethod``.
    """
    return sorted([member.__name__ for member in members])


def _write_datamodel_index_doc(datamodels: list, mode: str):
    """Write RST file for each datamodel.

    Parameters
    ----------
    datamodels: list
        Meshing and Solver datamodels.
    mode: str
        Fluent session mode either ``meshing`` or ``solver``.
    """
    full_folder_path = _get_docdir(mode, "", True)
    Path(full_folder_path).mkdir(parents=True, exist_ok=True)
    folder = Path(full_folder_path)
    index_file = folder / f"{folder.name}_contents.rst"
    datamodel_mode = f"{mode}.datamodel"
    with open(index_file, "w", encoding="utf8") as f:
        f.write(f".. _ref_{mode}_datamodel:\n\n")
        f.write(f"{datamodel_mode}\n")
        f.write(f"{'=' * len(datamodel_mode)}\n\n")
        f.write(".. automodule:: ansys.fluent.core.datamodel\n")
        f.write("   :autosummary:\n\n")
        f.write("   :autosummary-members:\n\n")
        f.write(".. toctree::\n")
        f.write("   :hidden:\n\n")
        for datamodel in datamodels:
            f.write(f"   {datamodel}/{datamodel}_contents\n")


def _get_reference(menu: type, menu_path: str, mode: str, is_datamodel: bool):
    """Get reference for RST file.

    Parameters
    ----------
    menu: type
        TUIMenu, TUIMethod
    menu_path: str
        Menu folder path.
    mode: str
        Fluent session mode either ``meshing`` or ``solver``.
    is_datamodel: bool
        Whether the menu is of datamodel.
    """
    if is_datamodel:
        if menu["name"].__name__ == "Root":
            reference = f".. _ref_{mode}_datamodel_{menu_path}:\n\n"
        else:
            reference = f".. _ref_{mode}_datamodel_{menu_path.rstrip('/').replace('/', '_')}:\n\n"
    else:
        if menu["name"].__name__ == "main_menu":
            reference = f".. _ref_{mode}_tui:\n\n"
        else:
            reference = f".. _ref_{mode}_tui_{menu_path.replace('/', '_')}:\n\n"
    return reference


def _get_title(mode: str, menu_path: str, menu: type, is_datamodel: bool):
    """Get title for RST.

    Parameters
    ----------
    mode: str
        Fluent session mode either ``meshing`` or ``solver``.
    menu_path: str
        Menu folder path.
    menu: type
        TUIMenu, TUIMethod
    is_datamodel: bool
        Whether the menu is of datamodel.
    """
    if is_datamodel:
        title = (
            f"{mode}.datamodel.{menu_path}"
            if menu["name"].__name__ == "Root"
            else menu["name"].__name__
        )
    else:
        title = (
            f"{mode}.tui"
            if menu["name"].__name__ == "main_menu"
            else menu["name"].__name__
        )
    return title


def _write_doc(menu: type, mode: str, is_datamodel: bool):
    """Write RST file for each menu.

    Parameters
    ----------
    menu: type
        TUIMenu, TUIMethod
    mode: str
        Fluent session mode either ``meshing`` or ``solver``.
    is_datamodel: bool
        Whether to generate datamodel RST files.
    """
    menu_name, menu_path = _get_menu_name_path(menu["name"], is_datamodel)
    full_folder_path = _get_docdir(mode, menu_path, is_datamodel)
    Path(full_folder_path).mkdir(parents=True, exist_ok=True)
    folder = Path(full_folder_path)
    index_file = folder / f"{folder.name}_contents.rst"
    title = _get_title(mode, menu_path, menu, is_datamodel)
    with open(index_file, "w", encoding="utf8") as f:
        f.write(_get_reference(menu, menu_path, mode, is_datamodel))
        f.write(f"{title}\n")
        f.write(f"{'=' * len(title)}\n\n")
        f.write(f".. autoclass:: {menu_name}\n")
        f.write(
            f"   :members: {', '.join(_get_sorted_members(menu['without_members']))}\n"
        )
        f.write("   :show-inheritance:\n")
        f.write("   :undoc-members:\n")
        f.write("   :exclude-members: __weakref__, __dict__\n")
        f.write("   :special-members: __init__\n")
        f.write("   :autosummary:\n")
        f.write("   :autosummary-members:\n\n")

        if menu["with_members"]:
            f.write(".. toctree::\n")
            f.write("   :hidden:\n\n")
            for member in _get_sorted_members(menu["with_members"]):
                f.write(f"   {member}/{member}_contents\n")


def _generate_all_attribute_classes(all_menus: list, main_menu: type):
    """Store all attribute classes with and without members into ``all_menus`` list.

    Parameters
    ----------
    all_menus: list
        ``all_menus`` list.
    main_menu: type
        ``main_menu`` or ``Root`` class.
    """
    with_members, without_members = _get_attribute_classes_with_and_without_members(
        main_menu
    )
    all_menus.append(
        dict(name=main_menu, with_members=with_members, without_members=without_members)
    )
    for member in with_members:
        _get_attribute_classes_recursively(member, all_menus)


def _generate_doc(all_menus: list, mode: str, is_datamodel: bool):
    """Write RST file for each attribute class.

    Parameters
    ----------
    all_menus: list
        ``all_menus`` list.
    mode: str
        Fluent session mode. Options are ``meshing`` and ``solver``.
    is_datamodel: bool
        Whether to generate datamodel RST files.
    """
    for menu in all_menus:
        _write_doc(menu, mode, is_datamodel)


def generate(main_menu: type, mode: str, is_datamodel: bool):
    """Generate RST files.

    Parameters
    ----------
    main_menu: type
        ``main_menu`` or ``Root`` class.
    mode: str
        Fluent session mode. Options are ``meshing`` and ``solver``.
    is_datamodel: bool
        Whether to generate datamodel RST files.
    """

    all_menus = []
    _generate_all_attribute_classes(all_menus, main_menu)
    _generate_doc(all_menus, mode, is_datamodel)
